//==========================================================================================================================
// Some basic USART routines for ATmega644
//==========================================================================================================================
#include <avr/io.h>

#include "serial.h"


// ------ config USART ------
void inline us_setup(uint16_t div)
{
	// rx, tx, U2X
	UCSR0A=1<<U2X0;
	UCSR0B=1<<RXEN0|1<<TXEN0;
	UCSR0C=1<<UCSZ01|1<<UCSZ00;
	UBRR0=div;
}

// ------ clear rx fifo ------
void inline rx_flush(void)
{
	UDR0;
	UDR0;
	UDR0;
}

// ------ get byte ------
uint8_t rx_byte(void)
{
	loop_until_bit_is_set(UCSR0A,RXC0);
	return(UDR0);
}

// ------ get word ------
uint16_t rx_word(void)
{
	union{
		uint8_t byte[2];
		uint32_t wrd;
	}wrd;
	wrd.byte[0]=rx_byte();
	wrd.byte[1]=rx_byte();
	return(wrd.wrd);
}

// ------ get dword ------
uint32_t rx_dword(void)
{
	union{
		uint8_t byte[4];
		uint32_t dwrd;
	}dwrd;
	dwrd.byte[0]=rx_byte();
	dwrd.byte[1]=rx_byte();
	dwrd.byte[2]=rx_byte();
	dwrd.byte[3]=rx_byte();
	return(dwrd.dwrd);
}





// ------ send byte ------
void tx_byte(uint8_t byte)
{
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=byte;
}

// ------ send word ------
void tx_word(uint16_t wrd)
{
	union{
		uint8_t byte[2];
		uint16_t wrd;
	}uni;
	uni.wrd=wrd;

	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[0];
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[1];
}

// ------ send dword ------
void tx_dword(uint32_t dwrd)
{
	union{
		uint8_t byte[4];
		uint32_t dwrd;
	}uni;
	uni.dwrd=dwrd;

	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[0];
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[1];
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[2];
	loop_until_bit_is_set(UCSR0A,UDRE0);
	UDR0=uni.byte[3];
}



