//==========================================================================================================================
// AVR I2S audio output demo - I2S functions
// Version V1.00, 10.2.2012
// 
// (C) 2012, Stanislav Maslan, s.maslan@seznam.cz
// www: http://www.elektronika.kvalitne.cz/ATMEL/necoteorie/tutorial/AudioTest/I2Semulation/I2S_emulation.html
//==========================================================================================================================


#include <avr/io.h>
#include <util/atomic.h>
#include <string.h>

#include "sound.h"
#include "main.h"


// I2S FIFO globals
uint8_t FIFO[FIFOsize];
uint8_t FIFOpref;
volatile uint8_t *FIFOrd;
uint8_t *FIFOwr;


// ------ set sample rate ------
void I2S_config(uint16_t srate)
{	
	// timer0 setup
	if(srate>F_CPU>>(3+8))
	{
		// fits into /8 divider range
		TCCR0B=1<<CS01;
		OCR0A=(F_CPU>>3)/srate-1;
	}
	else
	{
		// fits to /64 divider range
		TCCR0B=1<<CS01|1<<CS00;
		OCR0A=(F_CPU>>6)/srate-1;
	}
	TCCR0A=1<<WGM01;

	// SPI setup
	SPCR=1<<SPE|1<<MSTR;
	SPSR=1<<SPI2X;

}


// ------ set FIFO read pointer ------
void I2S_set_read(uint8_t *rd)
{
	// rather atomic
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
	{
		// set new read pointer
		FIFOrd=rd;
		// prefetch first byte
		FIFOpref=*rd;
	}
}


// ------ fill output FIFO asynchronously ------
inline void I2S_clear(void)
{
	memset((void*)FIFO,0x00,FIFOsize);
}


// ------ fill output fifo synchronously ------
void I2S_clear_sync(void)
{
	// leave if output stopped
	if(!(TIMSK0&1<<OCIE1A))
		return;

	uint8_t *wr=FIFOwr;
	uint16_t i;
	for(i=0;i<FIFOsize<<1;i++)
	{
		// wait for at least 1 free sample
		
		int16_t spc;
		do{
			volatile uint8_t *rdtmp;
			ATOMIC_BLOCK(ATOMIC_FORCEON)
    		{
				rdtmp=FIFOrd;
			}
			spc=rdtmp-wr;
			if(spc<0)
				spc+=FIFOsize;
		}while(spc<4);

		// write zero stereo sample
		*wr++=0;		

		// reset pointer if overrun
		if(wr>=(unsigned)FIFOsize+FIFO)
			wr=FIFO;
	}
	FIFOwr=wr;
}
