//==========================================================================================================================
// AVR I2S audio output demo - main I2S emmulation routine
// Version V1.00, 10.2.2012
// 
// (C) 2012, Stanislav Maslan, s.maslan@seznam.cz
// www: http://www.elektronika.kvalitne.cz/ATMEL/necoteorie/tutorial/AudioTest/I2Semulation/I2S_emulation.html
//==========================================================================================================================


#define __SFR_OFFSET 0
#include <avr/io.h>
#include "sound.h"
#include "main.h"

#include "macros.inc"

	
	;--- delay for N cycles ---
	.macro	del rr,cyc
	.if	\cyc==1
	nop
	.elseif	\cyc==2
	adiw	R30,0
	.elseif	\cyc==3
	nop
	adiw	R30,0
	.elseif	\cyc==4
	adiw	R30,0
	adiw	R30,0
	.elseif	\cyc==5
	adiw	R30,0
	adiw	R30,0
	nop
	.else
	ldi	\rr,\cyc/3
	dec	\rr
	brne	.-4
	.if	\cyc%3==1
	nop
	.elseif	\cyc%3==2
	adiw	R30,0
	.endif
	.endif
	.endm


	.global FIFO
	.global FIFOrd
	.global FIFOpref
	.global TIMER0_COMPA_vect

;************************************************************************************************
; I2S emmulation routine. Everytime it readsout one stereo sample from FIFO (FIFOrd pointer)
; and sends it on bus. To save some time I'm using unused i/o registers instead of push/pop.
; Don't try to optimize it - I'm pretty sure it can't be faster that it is now.         
;************************************************************************************************	
TIMER0_COMPA_vect:
	.if	DAC==1543
	;====== I2S version (TDA1543) ======
	out	OCR0B,R16	;push
	;
	lds	R16,FIFOpref	;get prefetched byte
	out	SPDR,R16	;high byte left
	;15 cycles pause
	in	R16,SREG
	out	GPIOR0,R16	;push
	out	GPIOR1,R30	;push
	out	GPIOR2,R31	;push
	ldsw	R30,FIFOrd	;FIFO read pointer
	adiw	R30,1
	del	R16,5
	;
	ld	R16,Z+
	out	SPDR,R16	;low byte left
	;12 cycles pause
	del	R16,9
	cpiwh	R30,R16,FIFO+FIFOsize-2
	;
	sbi	PORTB,DWS	;toggle WS
	nop
	ld	R16,Z+
	out	SPDR,R16	;high byte right
	ld	R16,Z+
	;15 cycles pause
	brne	PC+(1+2+1)*2
	ldiw	R30,FIFO	;reset read pointer if FIFO overflow
	rjmp	PC+(1+1+1)*2
	adiw	R30,0	;2x nop
	nop
	stsw	FIFOrd,R30	;store FIFO read pointer	
	ld	R31,Z	;prefetch next byte
	sts	FIFOpref,R31	;store it	
	del	R16,2
	;
	out	SPDR,R16	;low byte right
	;12 cycles pause
	del	R16,7
	in	R16,GPIOR0	;pop
	out	SREG,R16
	in	R31,GPIOR2	;pop
	in	R30,GPIOR1	;pop
	in	R16,OCR0B	;pop
	;
	cbi	PORTB,DWS	;toggle WS
	
	


	.elseif	DAC==1545
	;====== Japaneese two's complement version (TDA1545) ======
	out	OCR0B,R16	;push
	;
	lds	R16,FIFOpref	;get prefetched byte
	out	SPDR,R16	;high byte left
	;15 cycles pause
	in	R16,SREG
	out	GPIOR0,R16	;push
	out	GPIOR1,R30	;push
	out	GPIOR2,R31	;push
	ldsw	R30,FIFOrd	;FIFO read pointer
	adiw	R30,1
	del	R16,5
	;
	ld	R16,Z+
	out	SPDR,R16	;low byte left
	;15 cycles pause
	del	R16,10
	cpiwh	R30,16,FIFO+FIFOsize-2
	;
	ld	R16,Z+
	sbi	PORTB,DWS	;negate WS
	out	SPDR,R16	;high byte right
	ld	R16,Z+
	;15 cycles pause
	brne	PC+(1+2+1)*2
	ldiw	R30,FIFO	;reset FIFO read pointer if overflow
	rjmp	PC+(1+1+1)*2
	adiw	R30,0	;2x nop
	nop
	stsw	FIFOrd,R30	;store FIFO read pointer
	ld	R31,Z	;prefetch next byte
	sts	FIFOpref,R31	;store it	
	del	R16,2
	;
	out	SPDR,R16	;low byte right
	;15 cycles pause
	del	R16,10
	;
	in	R16,GPIOR0	;pop
	out	SREG,R16
	in	R31,GPIOR2	;pop
	in	R30,GPIOR1	;pop
	in	R16,OCR0B	;pop
	;
	cbi	PORTB,DWS	;negate WS
	
	.else
	.error "Invalid DAC type selection"
	.endif
		
	reti


