;compiler: avrasm
;------------------------------------------------------------------------------------------------------------------------
; Radic grafickeho LCD s budici ST8024+ST8016, zpracovava i data z touchscreenu
; E-mail: s.maslan@seznam.cz
;
; fusebity (0=naprogramovano jako v datasheetu):
; low:      0b11111111
; high:     0b11000001
; extended: 0b11111111
;
; XTAL=16MHz
; Externi RAM 70ns max!!! Idealne 55ns.
;------------------------------------------------------------------------------------------------------------------------ 

	.include "m64def.inc"
	;
	.dseg
	.org	0x0200	;interni RAM
RDaddress:	.byte	2	;zacatek vybraneho obrazoveho bufferu pro RD ukazatel
WRaddress:	.byte	2	;zacatek vybraneho obrazoveho bufferu pro WR ukazatel
TouchPadADx:	.byte	2	;primo zmerene souradnice dotyku X
TouchPadADy:	.byte	2	;primo zmerene souradnice dotyku Y
TPxVal:	.byte	6	;offset 0x00: korigovane souradnice dotyku X po filtraci
			;offset 0x02: korigovane souradnice dotyku X
			;offset 0x04: korigovane souradnice dotyku X posledni stisk (vyzivano filtrem)
TPyVal:	.byte	6	;offset 0x00: korigovane souradnice dotyku Y po filtraci
			;offset 0x02: korigovane souradnice dotyku Y
			;offset 0x04: korigovane souradnice dotyku Y posledni stisk (vyzivano filtrem)
TPxFiltrCTR:	.byte	1	;pocitadlo pro filtr pro X
TPyFiltrCTR:	.byte	1	;pocitadlo pro filtr pro Y
TPfiltrCycles:	.byte	1	;minimalni pocet mereni v toleranci pro pruchod filtrem
TPfiltrMaxDiff:	.byte	1	;max odchylka mezi jednotlivymi merenimi filtru
TPcalConst:	.byte	8	;kalibracni konstanty touchscreenu
TPfiltrDelay:	.byte	1	;zpozdeni vymazani souranic stisku po uvolneni dotyku
TPfiltrDelayReg:	.byte	1	;pocitadlo pro zpozdeni
	
	.org	0x1100	;externi RAM (doufam)
	;interface externi RAM je zapojen tak, ze nejnizsi bit adresy vybira pri cteni 
	;horni/dolni 4bity dat pro LCD data port
	;zadne cteni externi RAM mimo odesilani dat do LCD se NESMI provadet
	;zapis je mozny kdykoliv, inkrementaci adresy je treba provadet po 2B
ImgRAMstartI1:	.byte	19200	;obrazova data v RAM snimek 1
ImgRAMstartI2:	.byte	19200	;obrazova data v RAM snimek 2
ImgRAMstartI3:	.byte	19200	;obrazova data v RAM snimek 3


	.equ	XTAL=16000000	;frekvence krystalu

	.cseg
	.org	0x0000
	rjmp	init
	.org	0x0012
	rjmp	LineDelayTO

	;timeout pro casovani zpozdeni mezi radky
LineDelayTO:	set
	reti
	
	;zacatek programu
Init:	ldi	R16,0b00000000
	out	ddra,R16
	ldi	R16,0b00010101
	out	ddrb,R16
	ldi	R16,0b00000000
	out	ddrc,R16
	ldi	R16,0b00000000
	out	ddrd,R16
	ldi	R16,0b11111100
	out	ddre,R16
	ldi	R16,0b00000000
	sts	ddrf,R16
	ldi	R16,0b00000000
	sts	ddrg,R16
	;
	ldi	R16,0b11111111
	out	porta,R16
	out	portc,R16
	out	portd,R16
	ldi	R16,0b00011111
	;out	porte,R16
	ldi	R16,0b11111110
	;out	portb,R16
	;stack
	ldi	R16,high(0x1FF)
	out	sph,R16	;nastav SP
	ldi	R16,low(0x1FF)
	out	spl,R16	;nastav SP
	;aktivace externi RAM - zadny wait cykl, bez bus keepru, cely portc na adresy
	ldi	R16,0b10000000
	out	mcucr,R16
	ldi	R16,0b00000000
	sts	xmcra,R16
	ldi	R16,0b00000000
	sts	xmcrb,R16
	;aktivace AD prevodovniku
	ldi	R16,0b01000000
	out	admux,R16
	ldi	R16,0b10000110
	out	adcsra,R16
	ldi	R16,0b00000000
	sts	adcsrb,R16
	;1. prevod (pac trva nejak dyl)
	sbi	adcsra,adsc	;zacit prevod
WaitAD1stCok:	sbic	adcsra,adsc
	rjmp	WaitAD1stCok	;cekat na dokonceni konverze
	in	R16,adcl
	in	R16,adch
	
	;pocatecni stav pinu
	cbi	porte,4	;CL1
	cbi	porte,2	;FLM
	sbi	portb,2	;D-off
	sbi	portb,4	;CLK data in
	;M signal
	ldi	R16,0b01000000
	sts	tccr3a,R16
	ldi	R16,0b00001100
	sts	tccr3b,R16
	.equ	Mfreq=150
	ldi	R18,low(Mfreq)
	ldi	R19,high(Mfreq)
	rcall	IsetMfreq
	
	;timer 2 pro timeout cekani na data
	ldi	R16,0b00001010
	out	tccr2,R16	;XTAL/8, CTC
	ldi	R16,0b10000000
	out	timsk,R16	;INT CTC
	ldi	R16,100
	out	ocr2,R16	;cca 800 tiku perioda
	
	;nastavit kontrast
	ldi	R18,8
	rcall	IsetContrast
	
	;defaultni konstanty pro filtr touchscreenu
	ldi	R16,low(80)
	ldi	R17,high(80)
	sts	TPcalConst+0,R16
	sts	TPcalConst+1,R17
	sts	TPcalConst+2,R16
	sts	TPcalConst+3,R17
	ldi	R16,low(255)
	ldi	R17,high(255)
	sts	TPcalConst+4,R16
	sts	TPcalConst+5,R17
	ldi	R16,low(255)
	ldi	R17,high(255)
	sts	TPcalConst+6,R16
	sts	TPcalConst+7,R17
	ldi	R16,0
	sts	TPxFiltrCTR,R16
	sts	TPyFiltrCTR,R16	;nulovat pocitadla pro filtr
	ldi	R16,5
	sts	TPfiltrCycles,R16	;minimum cyklu v toleranci
	ldi	R16,3
	sts	TPfiltrMaxDiff,R16	;maximalni odchylka sousednich mereni (+/-)
	ldi	R16,5
	sts	TPfiltrDelay,R16	;zpozdeni vymazani souranic stisku po uvolneni dotyku
	sts	TPfiltrDelayReg,R16
	
	ldi	R16,low(ImgRAMstartI1)
	ldi	R17,high(ImgRAMstartI1)
	sts	RDaddress+0,R16
	sts	RDaddress+1,R17
	sts	WRaddress+0,R16
	sts	WRaddress+1,R17
	;hlavni smycka -  neustale refresh LCD, po kazdem radku 800tiku na prijem novych dat 
FRMsendNextFRM:	lds	R26,RDaddress+0
	lds	R27,RDaddress+1	;zacatek obrazku v RAM
	;
	ldi	R21,239	;pocet radku - 1
	;odeslat prvni radek
	ldi	R20,5	;5x16 slov do LCD X budice
FRMsend1stLn:	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	dec	R20
	brne	FRMsend1stLn
	;snimkovy synchronizacni tik
	sbi	porte,2	;FLM hi - zacatek framu
	nop
	nop
	sbi	porte,4
	cbi	porte,4	;radek tik
	nop
	cbi	porte,2	;FLM low - konec snimkovyh pulzu

	;odesilani jednotlivych radku (zbyvajicich 239)
FRMsendNextLn:	;tady se ceka 800tiku na vstupni data
	clr	R16
	out	tcnt2,R16	;nulovat timeout
	ldi	R16,0b10000000
	out	tifr,R16	;mazat priznak preruseni
	clt		;nulovat priznak timeoutu
	sei		;povolit preruseni
	
	cpi	R21,9	;provadet mereni napeti touchpadu?
	brcc	NotTouchMeasure
	;merit
	rcall	TouchPadMeasure
	;nemerit
NotTouchMeasure:	cpi	R21,129
	brcc	NotTouchMeasur2
	cpi	R21,120
	brcs	NotTouchMeasur2
	;merit
	push	R21
	subi	R21,120
	rcall	TouchPadMeasure
	pop	R21
	;nemerit
NotTouchMeasur2:
	
	sbis	pinb,5	;instrukce bo data?
	;instrukce
	rjmp	WaitInstrIn
	;data
WaitDataInP1:	brts	WaitDataTOrO	;timeout?
	;timeout ne, zadost o zapis dat?
	sbic	pinb,3
	;zadost o zapis ne, takze repete
	rjmp	WaitDataInP1
	;je zadost o ulozeni dat, ale neni to uz instrukce?
	sbis	pinb,5
	;instrukce
	rjmp	WaitDataTOrO
	;data
	in	R16,pind	;sosnout data byte
	cbi	portb,4	;potvrdit to MASTER obvodu
	st	Y,R16	;ulozit data
	adiw	R28,2	;inkrementovat adresu o 2B, pac nejnizsi bit adresy vybira nibble pro LCD
WaitDataInP2:	sbis	pinb,3	;cekat na dalsi data od MASTER obvodu
	;date ne, takze repete
	rjmp	WaitDataInP2
	;data jsou
	in	R16,pind	;sosnout data byte
	sbi	portb,4	;potvrdit to MASTER obvodu
	st	Y,R16	;ulozit data
	adiw	R28,2	;inkrementovat adresu o 2B, pac nejnizsi bit adresy vybira nibble pro LCD
	rjmp	WaitDataInP1	;a repete
	;timeout vyprsel, vosefovat konec RAM
WaitDataTOrO:	brtc	WaitDataTOrO	;cekat na time out
	cli		;zablokovat preruseni
	
	;ted uz odesilat dalsi radek
	ldi	R20,5	;5x16 slov do LCD X budice
FRMsendRemLn1:	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	ld	R16,X+
	dec	R20
	brne	FRMsendRemLn1
	
	sbi	porte,4
	cbi	porte,4	;radek pulz
	
	dec	R21	;radek - 1
	breq	FRMsendNextLnN
	rjmp	FRMsendNextLn	;dalsi radek
FRMsendNextLnN:	;vsechny radky vykresleny
	;ted vosefovat pripadne instrukce prijmute behem snimku
	; zatim nic tu
	;tady se ceka 800tiku po poslednim radku
	clr	R16
	out	tcnt2,R16	;nulovat timeout
	ldi	R16,0b10000000
	out	tifr,R16	;mazat priznak preruseni
	clt		;nulovat priznak timeoutu
	sei		;povolit preruseni
WaitOnLastLn:	brtc	WaitOnLastLn
	cli		;zakazat preruseni
		
	rjmp	FRMsendNextFRM	;zacni dalsi dalsi snimek
	
	
	;zpracovani instrukce
WaitInstrIn:	brts	WaitInstrInTO
	sbic	pinb,3
	;neni zadost o instrukci
	rjmp	WaitInstrIn
	;je zadost o zapis instrukce
	in	R16,pind	;sosnout data byte
	cbi	portb,4	;potvrdit prijem
WaitInstrA1:	sbis	pinb,3
	rjmp	WaitInstrA1	;cekat na dalsi byte
	in	R17,pind	;sosnout data byte
	sbi	portb,4	;potvrdit prijem
WaitInstrA2:	sbic	pinb,3
	rjmp	WaitInstrA2	;cekat na dalsi byte
	in	R18,pind	;sosnout data byte
	cbi	portb,4	;potvrdit prijem
WaitInstrA3:	sbis	pinb,3
	rjmp	WaitInstrA3	;cekat na posledni byte
	in	R19,pind	;sosnout data byte
	sbi	portb,4	;potvrdit prijem
	;R16:R17:R18:R19 obsahuje instrukci
	ldi	R30,low(InstrTable*2)
	ldi	R31,high(InstrTable*2)	;tabulka instrukci
	clr	R0
	cpi	R16,0x1D	;pocet existujicich instrukci
	brcs	InstrCanBe
	;nexistuje, nic nedelat
	rjmp	WaitInstrInTO
	;existuje
InstrCanBe:	lsl	R16
	rol	R0
	add	R30,R16
	adc	R31,R0	;vybrana instrukce
	lpm	R16,Z+
	lpm	R17,Z+
	movw	R30,R16
	icall		;vykonat instrukci
	;timeout, bo konec
WaitInstrInTO:	rjmp	WaitDataTOrO
	
	;odeslat word R16:R17
ItxWord:	ldi	R18,0b11111111
	out	ddrd,R18
	out	portd,R16
	cbi	portb,4
ItxWordACK1:	sbic	pinb,3
	rjmp	ItxWordACK1
	out	portd,R17
	sbi	portb,4
ItxWordACK2:	sbis	pinb,3
	rjmp	ItxWordACK2
	ldi	R18,0b00000000
	out	ddrd,R18
	cbi	portb,4
ItxWordACK3:	sbic	pinb,3
	rjmp	ItxWordACK3
	sbi	portb,4
ItxWordACK4:	sbis	pinb,3
	rjmp	ItxWordACK4
	ret
	
	;tabulka instrukci, kazdy word obsahuje adresu rutiny, ktera se ma vykonat
InstrTable:	.dw	IsetImgBuffWRa	;0x00 - nastavi WR adresu v obrazovem bufferu
	.dw	IsetImgBuff2RD	;0x01 - vybere obrazovy buffer pro zobarzovani
	.dw	IsetImgBuff2WR	;0x02 - vybere obrazovy buffer pro zapis
	.dw	Iblank	;0x03 - N/A
	.dw	Iblank	;0x04 - N/A
	.dw	Iblank	;0x05 - N/A
	.dw	Iblank	;0x06 - N/A
	.dw	Iblank	;0x07 - N/A
	.dw	Iblank	;0x08 - N/A
	.dw	Iblank	;0x09 - N/A
	.dw	IsetMfreq	;0x0A - nastavi frekvenci na pinu M (vstup=130..1000Hz)
	.dw	Iblank	;0x0B - N/A
	.dw	IsetContrast	;0x0C - nastavi kontrast (0..15)
	.dw	Iblank	;0x0D - N/A
	.dw	Iblank	;0x0E - N/A
	.dw	Iblank	;0x0F - N/A
	.dw	ItxTPcorX	;0x10 - odesle souradnici X touchscreenu - po korekci
	.dw	ItxTPcorY	;0x11 - odesle souradnici Y touchscreenu - po korekci
	.dw	ItxTPcorXF	;0x12 - odesle souradnici X touchscreenu - po korekci a filtraci
	.dw	ItxTPcorYF	;0x13 - odesle souradnici Y touchscreenu - po korekci a filtraci
	.dw	ItxTPdirX	;0x14 - odesle souradnici X touchscreenu - primo merene napeti 0-1023
	.dw	ItxTPdirY	;0x15 - odesle souradnici X touchscreenu - primo merene napeti 0-1023
	.dw	IsetTPfCycles	;0x16 - nastavi min. pocet mereni v toleranci pro touchscreen
	.dw	IsetTPfMaxDiff	;0x17 - nastavi max odchylku 2 mereni pro touchscreen
	.dw	IsetCalX1	;0x18 - nastavi kalibracni konstantu displeje 1 pro X
	.dw	IsetCalY1	;0x19 - nastavi kalibracni konstantu displeje 1 pro Y
	.dw	IsetCalX2	;0x1A - nastavi kalibracni konstantu displeje 2 pro X
	.dw	IsetCalY2	;0x1B - nastavi kalibracni konstantu displeje 2 pro Y
	.dw	IsetTPfDelay	;0x1C - nastavi zpozdeni mazani souradnic stisku
	;neprirazena instrukce
Iblank:	ret	
	;nastavuje WR adresu v image bufferu
IsetImgBuffWRa:	movw	R28,R18	;nastav adresu
	lsl	R28
	rol	R29	;x2, kvuli zpusobu adresace ext. RAM
	lds	R16,WRaddress+0
	lds	R17,WRaddress+1
	add	R28,R16
	adc	R29,R17	;pricist ofset zacatku bufferu
	ret
	;vybere obrazovy buffer pro zobarzovani
IsetImgBuff2RD:	ldi	R30,low(ImgBuffersT*2)
	ldi	R31,high(ImgBuffersT*2)
	lsl	R18
	clr	R19
	add	R30,R18
	adc	R31,R19
	lpm	R16,Z+
	lpm	R17,Z+
	sts	RDaddress+0,R16
	sts	RDaddress+1,R17
	ret
	;vybere obrazovy buffer pro zapis
	;pred prvnim zapisem dat po teto instrukci je treba nastavit WR adresu!!!  
IsetImgBuff2WR:	ldi	R30,low(ImgBuffersT*2)
	ldi	R31,high(ImgBuffersT*2)
	lsl	R18
	clr	R19
	add	R30,R18
	adc	R31,R19
	lpm	R16,Z+
	lpm	R17,Z+
	sts	WRaddress+0,R16
	sts	WRaddress+1,R17
	ret
	;tabulka obrazovych bufferu
ImgBuffersT:	.dw	ImgRAMstartI1
	.dw	ImgRAMstartI2
	.dw	ImgRAMstartI3
	;nastavi frekvenci na pinu M (vstup=130..1000Hz)
IsetMfreq:	ldi	R16,low(130)
	ldi	R17,high(130)
	cp	R18,R16
	cpc	R19,R17
	brcc	IsetMfreqNLR
	;prekrocit spodni hranici
	movw	R18,R16
IsetMfreqNLR:	ldi	R16,low(1001)
	ldi	R17,high(1001)
	cp	R18,R16
	cpc	R19,R17
	brcs	IsetMfreqNHR
	;prekrocil horni hranici
	movw	R18,R16
	subi	R18,low(1)
	sbci	R19,high(1)
IsetMfreqNHR:	ldi	R16,low(XTAL/512)
	ldi	R17,high(XTAL/512)
	call	div16u	;vypocitat konstantu pro orc3a
	sts	ocr3ah,R17
	sts	ocr3al,R16	;nastavit
	ret
	;nastavi kontrast LCD
IsetContrast:	cpi	R18,16
	brcs	IsetContrastNM
	;pokud je hodnota mimo 0-15, nastavit stred 8
	ldi	R18,8
IsetContrastNM:	mov	R19,R18	;kontrast do R18 a R19
	in	R16,ddre	;nacist stav pine
	andi	R16,0b00011111	;nechat ostani piny beze zmeny
	swap	R18
	lsl	R18
	andi	R18,0b11100000	;nechat jen kontrast
	or	R16,R18	;hodit tam kontrast
	out	ddre,R16	;a ven s tim
	in	R16,ddrb	;ted duha cast to samy
	andi	R16,0b11111110
	lsr	R19
	lsr	R19
	lsr	R19
	andi	R19,0b00000001
	or	R16,R19
	out	ddrb,R16	;a ven s tim
	ret
	;odesle souradnici X touchscreenu - po korekci
ItxTPcorX:	lds	R16,TPxVal+2
	lds	R17,TPxVal+3
	rcall	ItxWord
	ret
	;odesle souradnici Y touchscreenu - po korekci
ItxTPcorY:	lds	R16,TPyVal+2
	lds	R17,TPyVal+3
	rcall	ItxWord
	ret
	;odesle souradnici X touchscreenu - po korekci a filtraci
ItxTPcorXF:	lds	R16,TPxVal+0
	lds	R17,TPxVal+1
	rcall	ItxWord
	ret
	;odesle souradnici Y touchscreenu - po korekci a filtraci
ItxTPcorYF:	lds	R16,TPyVal+0
	lds	R17,TPyVal+1
	rcall	ItxWord
	ret
	;odesle souradnici X touchscreenu - primo merene napeti 0-1023 
ItxTPdirX:	lds	R16,TouchPadADx+0
	lds	R17,TouchPadADx+1
	rcall	ItxWord
	ret
	;odesle souradnici Y touchscreenu - primo merne napeti 0-1023
ItxTPdirY:	lds	R16,TouchPadADy+0
	lds	R17,TouchPadADy+1
	rcall	ItxWord
	ret
	;nastavi min. pocet mereni v toleranci pro touchscreen 
IsetTPfCycles:	sts	TPfiltrCycles,R18
	ret
	;nastavi max odchylku 2 mereni pro touchscreen 
IsetTPfMaxDiff:	sts	TPfiltrMaxDiff,R18
	ret
	;nastavi kalibracni konstantu displeje 1 pro X
IsetCalX1:	sts	TPcalConst+0,R18
	sts	TPcalConst+1,R19
	ret
	;nastavi kalibracni konstantu displeje 1 pro Y
IsetCalY1:	sts	TPcalConst+2,R18
	sts	TPcalConst+3,R19
	ret
	;nastavi kalibracni konstantu displeje 2 pro X (volat az po CalX1!!!)
IsetCalX2:	mov	R17,R18
	mov	R18,R19
	lds	R19,TPcalConst+0
	lds	R20,TPcalConst+1
	clr	R16
	sub	R17,R19
	sbc	R18,R20
	lsr	R18
	ror	R17
	ror	R16
	lsr	R18
	ror	R17
	ror	R16
	ldi	R18,low(320)
	ldi	R19,high(320)
	rcall	div16u
	sts	TPcalConst+4,R16
	sts	TPcalConst+5,R17
	ret
	;nastavi kalibracni konstantu displeje 2 pro Y (volat az po CalY1!!!)
IsetCalY2:	mov	R17,R18
	mov	R18,R19
	lds	R19,TPcalConst+2
	lds	R20,TPcalConst+3
	clr	R16
	sub	R17,R19
	sbc	R18,R20
	lsr	R18
	ror	R17
	ror	R16
	lsr	R18
	ror	R17
	ror	R16
	ldi	R18,low(240)
	ldi	R19,high(240)
	rcall	div16u
	sts	TPcalConst+6,R16
	sts	TPcalConst+7,R17
	ret
	;nastavi zpozdeni mazani souradnic stisku
IsetTPfDelay:	sts	TPfiltrDelay,R18
	ret
	
	;rutina pro mereni souradnic touchscreenu, korekci na 320*240 a digitlani filtraci "zakmitu"
TouchPadMeasure:	cpi	R21,8
	brne	TPMcycle1
	;1. cykl - init. mereni Y napeti
	ldi	R16,0b11111111
	sts	ddrf,R16
	ldi	R16,0b00000011
	sts	portf,R16	;napeti na Y odpor, hiZ na X odpor
	ldi	R16,0b01000010
	out	admux,R16	;vybrat AIN2
	rjmp	TPMcycleOk
TPMcycle1:	cpi	R21,7
	brne	TPMcycle2
	;2. cykl - start konverze napeti na AIN0
	ldi	R16,0b11000011
	sts	ddrf,R16
	ldi	R16,100
TPMcy2dL:	dec	R16
	brne	TPMcy2dL
	sbi	adcsra,adsc	;zacit prevod	
	rjmp	TPMcycleOk
TPMcycle2:	cpi	R21,6
	brne	TPMcycle3
	;3. cykl - nic, cekat na dokonceni konverze
	rjmp	TPMcycleOk
TPMcycle3:	cpi	R21,5
	brne	TPMcycle4
	;4. cykl - nacist zmerenou hodnotu, init. mereni X napeti
	in	R16,adcl
	in	R17,adch
	sts	TouchPadADy+0,R16
	sts	TouchPadADy+1,R17	;ulozit hodnotu Y napeti
	ldi	R16,0b11111111
	sts	ddrf,R16
	ldi	R16,0b00110000
	sts	portf,R16	;napeti na X odpor, hiZ na Y odpor
	ldi	R16,0b01000000
	out	admux,R16	;vybrat AIN0
	rjmp	TPMcycleOk
TPMcycle4:	cpi	R21,4
	brne	TPMcycle5
	;5. cykl - zacit konverzi
	ldi	R16,0b00111100
	sts	ddrf,R16
	ldi	R16,100
TPMcy5dL:	dec	R16
	brne	TPMcy5dL
	sbi	adcsra,adsc	;zacit prevod
	rjmp	TPMcycleOk
TPMcycle5:	cpi	R21,3
	brne	TPMcycle6
	;6. cykl - cekat na dokonceni konverze
	rjmp	TPMcycleOk
TPMcycle6:	cpi	R21,2
	brne	TPMcycle7
	;7. cykl - nacist zmerenou hodnotu, odpojit vsecny napeti z portu
	in	R16,adcl
	in	R17,adch
	sts	TouchPadADx+0,R16
	sts	TouchPadADx+1,R17	;ulozit hodnotu Y napeti
	ldi	R16,0b00000000
	sts	ddrf,R16
	ldi	R16,0b00000000
	sts	portf,R16	;hiZ vsude
	rjmp	TPMcycleOk
TPMcycle7:	;8. cykl - korekce na skutecne souradnice
	push	R21
	lds	R20,TouchPadADx+0
	lds	R21,TouchPadADx+1
	cpi	R20,low(70)
	ldi	R16,high(70)	;min pro vyhodnoceni dotyku
	cpc	R21,R16
	brcc	TPxCor2OK
	ldi	R16,low(0xFFFF)
	ldi	R17,high(0xFFFF)
	rjmp	TPxCor2OKn
TPxCor2OK:	lds	R16,TPcalConst+0
	lds	R17,TPcalConst+1
	sub	R20,R16
	sbc	R21,R17
	brcc	TPxCor1OK
	clr	R20
	clr	R21
TPxCor1OK:	ldi	R22,low(64)
	ldi	R23,high(64)
	rcall	mul16u
	lds	R18,TPcalConst+4
	lds	R19,TPcalConst+5
	rcall	div16u
	cpi	R16,low(320)
	ldi	R18,high(320)
	cpc	R17,R18
	brcs	TPxCor2OKn
	ldi	R16,low(319)
	ldi	R17,high(319)
TPxCor2OKn:	sts	TPxVal+2,R16
	sts	TPxVal+3,R17
	;
	lds	R20,TouchPadADy+0
	lds	R21,TouchPadADy+1
	cpi	R20,low(70)
	ldi	R16,high(70)
	cpc	R21,R16
	brcc	TPyCor2OK
	ldi	R16,low(0xFFFF)
	ldi	R17,high(0xFFFF)
	rjmp	TPyCor2OKn
TPyCor2OK:	lds	R16,TPcalConst+2
	lds	R17,TPcalConst+3
	sub	R20,R16
	sbc	R21,R17
	brcc	TPyCor1OK
	clr	R20
	clr	R21
TPyCor1OK:	ldi	R22,low(64)
	ldi	R23,high(64)
	rcall	mul16u
	lds	R18,TPcalConst+6
	lds	R19,TPcalConst+7
	rcall	div16u
	cpi	R16,low(240)
	brcs	TPyCor2OKn
	ldi	R16,low(239)
	ldi	R17,high(239)
TPyCor2OKn:	sts	TPyVal+2,R16
	sts	TPyVal+3,R17
	;
	pop	R21
	;digitalni filtrace zakmitu
	;ted filtrovat X
	lds	R16,TPxVal+2
	lds	R17,TPxVal+3	;nova hodnota
	cpi	R17,0xFF
	brne	TPxNotOut1
	;nedotyka se
	sts	TPxVal+4,R16
	sts	TPxVal+5,R17	;posledni hodnota na nedotyk.
	clr	R20
	sts	TPxFiltrCTR,R20	;nulovat pocet hodnot v toleranci
	rjmp	TPxIsOut1
	;dotyka se
TPxNotOut1:	lds	R20,TPxFiltrCTR
	cpi	R20,0
	brne	TPxNotOut2
	;minule se nedotykal
	sts	TPxVal+4,R16
	sts	TPxVal+5,R17	;posledni hodnota na soucasnou
	inc	R20	;inkrementovat pocet hodnot v toleranci
	sts	TPxFiltrCTR,R20	;ulozit
	rjmp	TPxIsOut1
	;uz se aspon jednou dotkl
TPxNotOut2:	lds	R18,TPxVal+4
	lds	R19,TPxVal+5
	cp	R16,R18
	cpc	R17,R19
	brcc	TPxCpPos1
	;negativni pomer
	sub	R18,R16
	sbc	R19,R17
	movw	R16,R18
	rjmp	TPxCpNeg1
	;pozitivni pomer
TPxCpPos1:	sub	R16,R18
	sbc	R17,R19
	;delta poslednich 2 hodnot
TPxCpNeg1:	lds	R18,TPfiltrMaxDiff
	clr	R19
	cp	R16,R18
	cpc	R17,R19
	lds	R16,TPxVal+2
	lds	R17,TPxVal+3	;nova hodnota
	sts	TPxVal+4,R16
	sts	TPxVal+5,R17	;posledni hodnota na soucasnou
	brcs	TPxDeltaOk
	;mimo toleranci
	clr	R20	;nlovat pocet stejnych hodnot
	sts	TPxFiltrCTR,R20	;ulozit
	rjmp	TPxIsOut1
	;v toleranci
TPxDeltaOk:	inc	R20
TPxIsOut1:	lds	R19,TPfiltrCycles
	cp	R20,R19
	brcc	TPxCanOut
	;jeste ne na vystup
	lds	R18,TPfiltrDelayReg
	cpi	R18,0
	breq	TPxNCanOut1
	dec	R18
	sts	TPfiltrDelayReg,R18
	rjmp	TPxNCanOut
TPxNCanOut1:	ldi	R18,0xFF
	sts	TPxVal+0,R18
	sts	TPxVal+1,R18	;vystup mimo
	rjmp	TPxNCanOut
	;uz muze na vystup
TPxCanOut:	lds	R18,TPfiltrDelay
	sts	TPfiltrDelayReg,R18
	sts	TPxVal+0,R16
	sts	TPxVal+1,R17	;vystup ne mimo
	mov	R20,R19
TPxNCanOut:	sts	TPxFiltrCTR,R20	;ulozit
	
	;ted filtrovat Y
	lds	R16,TPyVal+2
	lds	R17,TPyVal+3	;nova hodnota
	cpi	R17,0xFF
	brne	TPyNotOut1
	;nedotyka se
	sts	TPyVal+4,R16
	sts	TPyVal+5,R17	;posledni hodnota na nedotyk.
	clr	R20
	sts	TPyFiltrCTR,R20	;nulovat pocet hodnot v toleranci
	rjmp	TPyIsOut1
	;dotyka se
TPyNotOut1:	lds	R20,TPyFiltrCTR
	cpi	R20,0
	brne	TPyNotOut2
	;minule se nedotykal
	sts	TPyVal+4,R16
	sts	TPyVal+5,R17	;posledni hodnota na soucasnou
	inc	R20	;inkrementovat pocet hodnot v toleranci
	sts	TPyFiltrCTR,R20	;ulozit
	rjmp	TPyIsOut1
	;uz se aspon jednou dotkl
TPyNotOut2:	lds	R18,TPyVal+4
	lds	R19,TPyVal+5
	cp	R16,R18
	cpc	R17,R19
	brcc	TPyCpPos1
	;negativni pomer
	sub	R18,R16
	sbc	R19,R17
	movw	R16,R18
	rjmp	TPyCpNeg1
	;pozitivni pomer
TPyCpPos1:	sub	R16,R18
	sbc	R17,R19
	;delta poslednich 2 hodnot
TPyCpNeg1:	lds	R18,TPfiltrMaxDiff
	clr	R19
	cp	R16,R18
	cpc	R17,R19
	lds	R16,TPyVal+2
	lds	R17,TPyVal+3	;nova hodnota
	sts	TPyVal+4,R16
	sts	TPyVal+5,R17	;posledni hodnota na soucasnou
	brcs	TPyDeltaOk
	;mimo toleranci
	clr	R20	;nlovat pocet stejnych hodnot
	sts	TPyFiltrCTR,R20	;ulozit
	rjmp	TPyIsOut1
	;v toleranci
TPyDeltaOk:	inc	R20
TPyIsOut1:	lds	R19,TPfiltrCycles
	cp	R20,R19
	brcc	TPyCanOut
	;jeste ne na vystup
	lds	R18,TPfiltrDelayReg
	cpi	R18,0
	breq	TPyNCanOut1
	dec	R18
	sts	TPfiltrDelayReg,R18
	rjmp	TPyNCanOut
TPyNCanOut1:	ldi	R18,0xFF
	sts	TPyVal+0,R18
	sts	TPyVal+1,R18	;vystup mimo
	rjmp	TPyNCanOut
	;uz muze na vystup
TPyCanOut:	lds	R18,TPfiltrDelay
	sts	TPfiltrDelayReg,R18
	sts	TPyVal+0,R16
	sts	TPyVal+1,R17	;vystup ne mimo
	mov	R20,R19
TPyNCanOut:	sts	TPyFiltrCTR,R20	;ulozit

TPMcycleOk:	ret
;------------------------------------------------------------------------------------------
; div16u - 16/16 bit unsigned division, max. speed optimized
; deleni dvou neznaminkovych 16-bit cisel, maximalni optimalizace na rychlost
;
; editor: Pavel Borovsky, Czech Republic, Europe 3.6.2000
; email: profesor@ti.cz
;
; dDd16r0u:dDd16r1u <- dDd16r0u:dDd16r1u div dDs16r0u:dDs16r1u
; dRe16r0u:dRe16r1u <- dDd16r0u:dDd16r1u mod dDs16r0u:dDs16r1u
;
; words:  142 + ret                        (puvodni od ATMELu: 196 + ret)
; cycles: 126/134/142 + ret [min/avg/max]  (puvodni od ATMELu: 148/172/196 + ret)
;------------------------------------------------------------------------------------------
	.def	dRe16r0u=R14
	.def	dRe16r1u=R15
	.def	dDd16r0u=R16
	.def	dDd16r1u=R17
	.def	dDs16r0u=R18
	.def	dDs16r1u=R19
	;
div16u:	clr	dRe16r0u
	sub	dRe16r1u,dRe16r1u
	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_1
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_1:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_2
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_2:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_3
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_3:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_4
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_4:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_5
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_5:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_6
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_6:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_7
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_7:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_8
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_8:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_9
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_9:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_10
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_10:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_11
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_11:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_12
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_12:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_13
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_13:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_14
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_14:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_15
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_15:	rol	dDd16r0u
	rol	dDd16r1u
	rol	dRe16r0u
	rol	dRe16r1u
	sub	dRe16r0u,dDs16r0u
	sbc	dRe16r1u,dDs16r1u
	brcc	d16u_16
	add	dRe16r0u,dDs16r0u
	adc	dRe16r1u,dDs16r1u
d16u_16:	rol	dDd16r0u
	rol	dDd16r1u
	com	dDd16r0u
	com	dDd16r1u
	ret
;------------------------------------------------------------------------------------------------------------------------
; mul16u - 16*16 bit unsigned multiplication
; nasobeni 16x16 unsigned
; R16:R17:R18:R19 <- R20:R21 * R22:R23
;
; words:  13 + ret
; cycles: 17 + ret
;------------------------------------------------------------------------------------------------------------------------
mul16u:	clr	R2
	mul	R23,R21
	movw	R18,R0
	mul	R22,R20
	movw	R16,R0
	mul	R23,R20
	add	R17,R0
	adc	R18,R1
	adc	R19,R2
	mul	R21,R22
	add	R17,R0
	adc	R18,R1
	adc	R19,R2
	ret