;compiler: avrasm
	.include "m32def.inc"
	.include "macros.asm"

	.equ	XTAL=16000000
	.equ	FIFOsize=1920	;audio FIFO size (multiple of 2 (or 4 for stereo)!!!)
		
	.dseg
FIFOreadP:	.byte	2	;FIFO read pointer
RXtoVar:	.byte	1	;timeout variable
SamplingRate:	.byte	2	;actual sampling rate
FlagVar:	.byte	1	;some bit flags variable
FIFObuffer:	.byte	FIFOsize	;FIFO buffer
	
	.cseg
	.org	0x0000
	rjmp	Init	;init
	.org	0x0004
	rjmp	INTrotEnc0	;rotary encoder 0 INT
	.org	0x0008
	rjmp	I2Swrite	;I2S DMA output
	.org	0x000E
	rjmp	RXtimeout	;receive timeout
	.org	0x0014
	rjmp	INTLCDDMArefresh	;LCD DMA refresh
	.org	0x001A
	rjmp	COMrxData	;DMA RX byte to audio FIFO
		
	
	;portd pins
	.equ	CTS=2	;HW flow ctrl
	.equ	RE0_ck=3	;rotary encoder 1 clock pin	
	.equ	RE0_dt=4	;rotary encoder 1 data pin
	
;************************************************************************************************************************
;	init
;************************************************************************************************************************	
Init:	outi	ddrc,0b00001111	;LCD data pins
	outi	ddrd,(1<<CTS)|(1<<LCK)|(1<<LRS)
	outi	portd,(1<<RE0_ck)|(1<<RE0_dt)	;rotary encoder pull-up resistors
	
	cbi	portd,CTS
	cbi	LCDportE,LCK

	outiw	sp,ramend	;sp to RAMend
	
	ldi	R16,50	;500ms
	rcall	DelayVar
		
	;USART
	outi	ucsra,(1<<u2x)
	outi	ucsrb,(1<<rxen)|(1<<txen)|(1<<rxcie)
	outiw	ubrr,0
	
	;timer 1 - receive timeout
	outi	tccr1a,0
	outi	tccr1b,(1<<ctc1)|(1<<cs10)|(1<<cs12)
	outiw	ocr1a,XTAL/1024/1-1	;T = 1s
	outi	timsk,(1<<ocie1a)
	
	;external IRQ for rotary encoder
	outi	mcucr,(1<<isc11)	;INT0 falling edge interruption (preruseni na sestupnou hranu na INT0)
	outi	gicr,(1<<int1)	;enable INT0 (povolit INT0 preruseni)
	
	;flush RX buffer
	in	R16,udr
	in	R16,udr
	
	
	ldi	R16,2
	rcall	Dinit	;init DMA LCD refresh
		
	rcall	I2Sinit	;init I2S emulator
	
	stsiw	SamplingRate,22050	;initial sampling rate
	stsi	FlagVar,0	;clear bit flags
		
	sei		;enable INT
	
	;reinit data stream
RXtoGo:	ldiw	R24,FIFOsize/2
	ldiw	R28,FIFObuffer
	ldi	R16,0	;clear FIFO
	st	Y+,R16
	st	Y+,R16
	dbnzwh	R24,PC-3
	;reset pointers
	ldiw	R26,FIFObuffer+0	;write pointer (don't use R26 anywhere!)
	stsiw	FIFOreadP,FIFObuffer+0	;read poniter
	clr	R5	;initial RX byte counter
	clr	R6	;PCM header test allowed 
	;reset timeout
	rcall	RedrawLCDinf2
	rcall	RXtoReset
	
	
;--------------------------------------------------------------------------------------------------------	
;	main play loop
;--------------------------------------------------------------------------------------------------------
	;free capacity in the FIFO?
Loop:	cli
	nop
	movw	R16,R26	;write pointer
	ldsw	R28,FIFOreadP	;read pointer
	subw	R28,R16
	brcc	FIFOcanWrite1
	subiw	R28,-FIFOsize	
FIFOcanWrite1:	sei		;Y bytes free
	cpiwh	R28,R16,32
	brsh	COMrxHandle1
	;CTS=1 (stop sending data)
	sbi	portd,CTS
	rjmp	COMrxHandle2
COMrxHandle1:	cpiwh	R28,R16,48
	brlo	COMrxHandle2
	;CTS=0 (start sending data)
	cbi	portd,CTS
COMrxHandle2:	;timeout test
	lds	R16,RXtoVar
	tst	R16
	breq	RXtoGo	;timeout - reset receiver
	;now RIFF header test for auto set sampling rate (mostly is works :-)
	tst	R6
	rjne	RIFFtNAllowed
	;test allowed, ? bytes received?
	ldi	R16,64
	cp	R5,R16
	rjlo	RIFFtNAllowed	;at least 64B in buffer?
	;64+ btes in FIFO buffer, read RIFF headder
	ldiw	R28,FIFObuffer
	lddd	R16,Y+0
	cpi	R16,'R'
	ldi	R21,'I'
	ldi	R22,'F'
	ldi	R23,'F'
	cpct	R17,R21
	brne	RIFFtNAllowed	;exit test - there is not RIFF header
	adiw	R28,8
	lddd	R16,Y+0
	cpi	R16,'W'
	ldi	R21,'A'
	ldi	R22,'V'
	ldi	R23,'E'
	cpct	R17,R21
	brne	RIFFtNAllowed	;exit test - there is not WAVE ID
	;now we know that it's RIFF structure and WAVE file
	;and we need to find "fmt " subchunk - there is info whether it's PCM WAV, smaple rate, ...
	adiw	R28,4	;to first subchunk
	ldi	R24,4	;4 tries to found our header
RIFFtryNextSC2:	lddd	R16,Y+0
	cpi	R16,'f'
	ldi	R21,'m'
	ldi	R22,'t'
	ldi	R23,' '
	cpct	R17,R21
	brne	RIFFtryNextSC1
	;"fmt " subchunk found, get sample rate and channels number
	adiw	R28,4
	ldd	R16,Y+6	;channels?
	sts	I2Schannels,R16	;set it
	ldd	R16,Y+8+0
	ldd	R17,Y+8+1	;get sample rate
	stsw	SamplingRate,R16	;store new value
	rcall	I2SsetSMPLrate	;set it
	rcall	RedrawLCDinf1	;rewrite LCD info
	rjmp	RIFFtNAllowed1	;exit
	;ID not found, try next 
RIFFtryNextSC1:	ldd	R16,Y+4+0	;only low byte of subchunk len
	clr	R17
	adiw	R28,8
	addw	R28,R16	;add to the pointer
	cpiwh	R28,R16,FIFObuffer+FIFOsize-128
	brsh	RIFFtNAllowed1	;pointer out of range
	dec	R24
	brne	RIFFtryNextSC2	;try again if was not last try
RIFFtNAllowed1:	inc	R6	;ban RIFF test in the future
RIFFtNAllowed:	;now test for manual sampling rate change
	lds	R16,FlagVar
	rjrc	R16,0,MCSRno
	;it was changed
	cli
	cbr	R16,1
	sts	FlagVar,R16
	ldsw	R16,SamplingRate	;load new sampling rate
	sei
	sbi	portd,CTS	;stop sendig new data
	rcall	I2SsetSMPLrate
	rcall	RedrawLCDinf1	;rewrite LCD info
MCSRno:	rjmp	Loop


;------------------------------------------------------------------------------------------------------------------------
;	write LCD info
;------------------------------------------------------------------------------------------------------------------------
RedrawLCDinf1:	rcall	DclrAll	;clear LCD
	ldiw	R30,TxtInf1T*2
	rcall	DwrStrFLASH	;write string
	cli
	ldsw	R16,SamplingRate	;get sampling rate value
	sei
	rcall	bin2BCD16p5	;to BCD
	rcall	DwrBCD16sz	;write to LCD without leading zeroes
	ldi	R16,'H'
	rcall	DwrChar
	ldi	R16,'z'
	rcall	DwrChar	;"Hz"
	ldi	R16,' '
	rcall	DwrChar
	lds	R16,I2Schannels
	cpi	R16,1
	brne	PC+4
	ldiw	R30,TxtMonoT*2
	rjmp	PC+3
	ldiw	R30,TxtStereoT*2
	rcall	DwrStrFLASH
	;
	rcall	Drefresh	;start DMA refresh
	ret
	
;------------------------------------------------------------------------------------------------------------------------
;	write on LCD timeout
;------------------------------------------------------------------------------------------------------------------------
RedrawLCDinf2:	rcall	DclrAll	;clear LCD
	ldiw	R30,TxtInf2T*2
	rcall	DwrStrFLASH	;wriet string
	rcall	Drefresh	;start DMA refresh
	ret
	
TxtInf1T:	.db	"Playing ...",0x00,"Fs=",0xFF
TxtInf2T:	.db	"Timeout ...",0xFF
TxtMonoT:	.db	"mono",0xFF
TxtStereoT:	.db	"stereo",0xFF


;************************************************************************************************************************
;	COM receive complete interruption routine, received byte are stored into audio FIFO
;	don't use X pointer and R5,R7,R8 register anywhere else (it's not pushed)!!!
;************************************************************************************************************************
COMrxData:	mov	R7,R16
	in	R8,sreg	;instead of push
	;store byte

	in	R16,udr	;get byte
	st	X+,R16	;store byte
	inc	R5	;bytes recevied++
	
	cpiwh	R26,R16,FIFObuffer+FIFOsize
	brlo	PC+3	;end of FIFO buffer?
	ldiw	R26,FIFObuffer+0	;yes, reset pointer
	
	
	;and again if something is in receive buffer
	sbis	ucsra,rxc
	rjmp	COMrxNoNext
	
	in	R16,udr	;get byte
	st	X+,R16	;store byte
	inc	R5	;bytes recevied++
	
	cpiwh	R26,R16,FIFObuffer+FIFOsize
	brlo	PC+3	;end of FIFO buffer?
	ldiw	R26,FIFObuffer+0	;yes, reset pointer
	
COMrxNoNext:	;
	outi	sfior,(1<<psr10)
	outiw	tcnt1,0
	outi	tifr,(1<<ocf1a)
	stsi	RXtoVar,0xFF
	;
	out	sreg,R8
	mov	R16,R7	;instead of pop
	reti



;************************************************************************************************************************
; I2S and "japanese audio format" emulation routines called by timer 2 compare match interruption
; each IRQ takes 1 16bit sample (or 2 in stereo mode) from FIFO and send it to the DAC (low byte first!)
; interruption can't be blocked by another routine for longer than 1Ts (it sounds very bad)
; there are some in/out instructions into some unused registers instead of push/pop (optimalization)
; wrote for internal RAM FIFO (2 cycle access, for external RAM must be rebuild)!!!
;
; rutiny pro emulaci I2S a "standardni japonske audio sbernice", vlastni rutina volana pres preruseni 
; timeru 2, pokazde vezme 1 (pro stereo 2) 16bit vzorky (low byte ulozen prvni!), posle na DAC
; preruseni nesmi byt blokovane dele nez asi 1Tvz jinak to zni dost spatne
; misto push/pop jsou tam in/out do nevyuzitych registru - usetri se par cyklu
; je to napsano pro FIFO v interni RAM (2 cykly na pristup, pro externi RAM se musi predelat prodlevy)!!!
;
; author: Stanislav Maslan, s.maslan@seznam.cz
; last modify: 8.5.2007  [dd.mm.yyyy :-)]
;************************************************************************************************************************
	.equ	DAC=1543	;select protocol: 1543=TDA1543 (I2S), 1545=TDA1545 ("japanese format")
	.equ	channels=0	;number of channels (1 or 2 for preprocessor selcetion or 0 for realtime change)
	.equ	smplrate=22050	;initial sampling rate in Hz
	
	.equ	WSport=portb	;WS pin port
	.equ	WSddr=ddrb	;WS pin ddr
	.equ	WSpin=4	;WS pin
	.equ	DTCKddr=ddrb	;data and clock port
	.equ	MOSIpin=5	;I2S data
	.equ	SCKpin=7	;I2S clock
	
	.dseg
	.if	channels==0
I2Schannels:	.byte	1	;number of channels for SW selectable mono/stereo
	.endif
	.cseg
;------------------------------------------------------------------------------------------------------------------------
;	initialization routine, call first in main init 
;------------------------------------------------------------------------------------------------------------------------
I2Sinit:	;set port directions
	sbrii	WSddr,(1<<WSpin)
	sbrii	DTCKddr,(1<<MOSIpin)|(1<<SCKpin)
	cbi	WSport,WSpin
	
	;timer 2 - I2S DMA output
	outi	tccr2,(1<<ctc2)|(1<<cs21)
	outi	ocr2,XTAL/8/smplrate-1	;sampling rate
	sbrii	timsk,(1<<ocie2)
	
	;SPI - I2S emulation at 8MHz
	outi	spcr,(1<<spe)|(1<<mstr)
	outi	spsr,(1<<spi2x)
		
	.if	channels==0
	stsi	I2Schannels,1	;default mono if SW selection activated
	.endif
	
	;reset FIFO read pointer
	stsiw	FIFOreadP,FIFObuffer+0
	ret
	
;------------------------------------------------------------------------------------------------------------------------
;	set sampling rate R16:R17 (recomended from 8 to 48kHz at XTAL=16MHz)
;------------------------------------------------------------------------------------------------------------------------
I2SsetSMPLrate:	;estimate ocr2 value (ocr2 = XTAL/8/R16:R17-1)
	mov	R19,R16
	mov	R20,R17
	clr	R21
	ldit	R16,XTAL/8
	rcall	div24uc3
	dec	R16
	out	ocr2,R16	;set new ocr2
	ret
	
;------------------------------------------------------------------------------------------------------------------------
;	main emulation routine (called from timer 2 compare match interruption)
;------------------------------------------------------------------------------------------------------------------------
I2Swrite:	out	ocr1bh,R16	;misto push
	out	ocr1bl,R30	;misto push
	in	R16,sreg	;misto push
	out	icr1h,R31	;misto push
	out	icr1l,R16	;misto push

	ldsw	R30,FIFOreadP	;load read FIFO address
	.if	DAC==1543
	
	.if	channels==0
	lds	R16,I2Schannels	;SW channels selector
	cpi	R16,1	;mono?
	breq	I2S_mono	;yes
	.endif
;------------------------------------------------------------------------------------------------------------------------
;	output for I2S DAC TDA1543 (selectable mono, stereo or realtime selectable stereo)
;------------------------------------------------------------------------------------------------------------------------
	;get left channel high byte
	ldd	R16,Z+1
	.if	(channels!=1) && (channels!=2) && (channels!=0)
	.error	"specify number of DAC channels (1 or 2)!!!"
	.endif
	out	spdr,R16	;TX it
	;13 cycles wait
	ldi	R16,4
	dec	R16
	brne	PC-1
	nop
	;now read left channel low byte
	ldd	R16,Z+0
	.if	(channels==1)
	adiw	R30,0	;stay on this sample
	.elif	(channels==2) || (channels==0)
	adiw	R30,2	;to next sample
	.else
	.error	"specify number of DAC channels (1 or 2)!!!"
	.endif
	out	spdr,R16	;TX it
	;12 cycles wait
	ldi	R16,3
	dec	R16
	brne	PC-1
	cpiwh	R30,R16,FIFObuffer+FIFOsize-2	;end of FIFO???
	;
	sbi	WSport,WSpin	;change WS state to the right channel
	nop
	ldd	R16,Z+1	;get right channel high byte
	out	spdr,R16	;TX it
	ldd	R16,Z+0	;get low byte
	;wait 13 cycles
	brlo	NeKonec
	;yes, it was FIFO end
	ldiw	R30,FIFObuffer	;reset pointer to the FIFO start
	rjmp	JoKonec
NeKonec:	adiw	R30,2
	nop
JoKOnec:	stsw	FIFOreadP,R30	;store read pointer back
	adiw	R30,0
	adiw	R30,0
	adiw	R30,0
	;
	out	spdr,R16	;TX richt channel low byte
	;wait 15 cycles
	adiw	R30,0	;2 cycles
	adiw	R30,0	;2 cycles
	adiw	R30,0	;2 cycles
	nop
	in	R31,icr1h	;pop
	in	R16,icr1l	;pop
	out	sreg,R16	;pop
	in	R16,ocr1bh	;pop
	in	R30,ocr1bl	;pop
	cbi	WSport,WSpin	;change WS stete back to the left channel
	reti

	.elif	DAC==1545
;------------------------------------------------------------------------------------------------------------------------
;	output for "japanese audio format" DAC TDA1545
;------------------------------------------------------------------------------------------------------------------------
	;get left channel high byte
	ldd	R16,Z+1
	.if	channels!=1 && channels!=2 && channels!=0
	.error	"specify number of DAC channels (1 or 2)!!!"
	.endif
	out	spdr,R16	;TX it
	;15 cycles wait
	ldi	R16,4
	dec	R16
	brne	PC-1
	nop
	;now read left channel low byte
	ldd	R16,Z+0
	.if	channels==1
	adiw	R30,0
	.elif	channels==2 || channels==0
	adiw	R30,2
	.else
	.error	"specify number of DAC channels (1 or 2)!!!"
	.endif
	out	spdr,R16	;TX it 
	;wait 15 cycles
	ldi	R16,4
	dec	R16
	brne	PC-1
	cpiwh	R30,R16,FIFObuffer+FIFOsize-2	;end of FIFO???
	;
	ldd	R16,Z+1	;get right channel high byte
	sbi	portb,2	;WS pin to the right channel
	out	spdr,R16	;TX right high
	ldd	R16,Z+0	;get right channel low byte
	;wait 15 cycles
	brne	NeKonec
	;end of FIFO
	ldiw	R30,FIFObuffer	;reset poitner
	rjmp	JoKonec
NeKonec:	adiw	R30,2
	nop
JoKOnec:	stsw	FIFOreadP,R30
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	;
	out	spdr,R16	;TX low byte
	;wait 15
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2C
	in	R31,icr1h	;pop
	in	R16,icr1l	;pop
	out	sreg,R16	;pop
	in	R16,ocr1bh	;pop
	in	R30,ocr1bl	;pop
	;
	cbi	WSport,WSpin	;change WS state back to left channel
	reti
	
	.else
	.error	"Specify DAC type (1545 od 1543)!!!"
	.endif
	


I2S_mono:	.if	channels==0 	;compile only if 0 channels selected
	.if	DAC==1543
;------------------------------------------------------------------------------------------------------------------------
;	output for I2S DAC TDA1543 (realtime selectable mono)
;------------------------------------------------------------------------------------------------------------------------
	;get left channel high byte
	ldd	R16,Z+1
	out	spdr,R16	;TX it
	;15 cycles wait
	ldi	R16,5
	dec	R16
	brne	PC-1
	;now read left channel low byte
	ldd	R16,Z+0
	out	spdr,R16	;TX it
	;12 cycles wait
	ldi	R16,3
	dec	R16
	brne	PC-1
	cpiwh	R30,R16,FIFObuffer+FIFOsize-2	;end of FIFO???
	;
	sbi	WSport,WSpin	;change WS state to the right channel
	nop
	ldd	R16,Z+1	;get right channel high byte
	out	spdr,R16	;TX it
	ldd	R16,Z+0	;get low byte
	;wait 15 cycles
	brlo	NeKonecM
	;yes, it was FIFO end
	ldiw	R30,FIFObuffer	;reset pointer to the FIFO start
	rjmp	JoKonecM
NeKonecM:	adiw	R30,2
	nop
JoKOnecM:	stsw	FIFOreadP,R30	;store read pointer back
	adiw	R30,0
	adiw	R30,0
	adiw	R30,0	;2 cycles
	;
	out	spdr,R16	;TX richt channel low byte
	;wait 15 cycles
	adiw	R30,0	;2 cycles
	adiw	R30,0	;2 cycles
	adiw	R30,0	;2 cycles
	nop
	in	R31,icr1h	;pop
	in	R16,icr1l	;pop
	out	sreg,R16	;pop
	in	R16,ocr1bh	;pop
	in	R30,ocr1bl	;pop
	cbi	WSport,WSpin	;change WS stete back to the left channel
	reti
	
	.elif	DAC==1545
;------------------------------------------------------------------------------------------------------------------------
;	output for "japanese audio format" DAC TDA1545 (realtime selectable mono)
;------------------------------------------------------------------------------------------------------------------------
	;get left channel high byte
	ldd	R16,Z+1
	out	spdr,R16	;TX it
	;15 cycles wait
	ldi	R16,5
	dec	R16
	brne	PC-1
	;now read left channel low byte
	ldd	R16,Z+0
	out	spdr,R16	;TX it 
	;wait 15 cycles
	ldi	R16,4
	dec	R16
	brne	PC-1
	cpiwh	R30,R16,FIFObuffer+FIFOsize-2	;end of FIFO???
	;
	ldd	R16,Z+1	;get right channel high byte
	sbi	portb,2	;WS pin to the right channel
	out	spdr,R16	;TX right high
	ldd	R16,Z+0	;get right channel low byte
	;wait 15 cycles
	brne	NeKonecM
	;end of FIFO
	ldiw	R30,FIFObuffer	;reset poitner
	rjmp	JoKonecM
NeKonecM:	adiw	R30,2
	nop
JoKOnecM:	stsw	FIFOreadP,R30
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	;
	out	spdr,R16	;TX low byte
	;wait 15
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2c
	adiw	R30,0	;2C
	in	R31,icr1h	;pop
	in	R16,icr1l	;pop
	out	sreg,R16	;pop
	in	R16,ocr1bh	;pop
	in	R30,ocr1bl	;pop
	;
	cbi	WSport,WSpin	;change WS state back to left channel
	reti	
	.else
	.error	"Specify DAC type (1545 od 1543)!!!"
	.endif
	.endif		


	
;************************************************************************************************************************
;	rotary encoder INT (preruseni rotacniho enkoderu)
;************************************************************************************************************************	
INTrotEnc0:	pushw	R16
	push	R18
	in	R16,sreg
	push	R16
	sei		;enable interruption - USART at 2Mbit/sec is very critical and can't be holded 
	;
	ldsw	R16,SamplingRate	;load our var (nacist nasi promennou)
	sbis	pind,RE0_dt	;decide whether inc/dec according to log.state on data pin ER0_dt
			;rozhodnout jestli pricitat nebo odecitat podle stavu data pinu RE_dt
	rjmp	INTrotEnc0_inc
	;inc var
	subiw	R16,-100	;inc
	cpiwh	R16,R18,48000
	brlo	PC+3
	ldiw	R16,48000	;max 48000Hz
	rjmp	INTrotEnc0_dec
	;dec var
INTrotEnc0_inc:	subiw	R16,100	;dec
	cpiwh	R16,R18,8000
	brsh	PC+3
	ldiw	R16,8000	;min 8000Hz	
INTrotEnc0_dec:	stsw	SamplingRate,R16	;store it back (a ulozit zmenenou hodnotu zpet)
	;now clear INT0 IRQ request if there were some fake requests
	;nulovat priznak preruseni pokud byly nejake vetsi zakmity
	outi	gifr,(1<<intf1)
	rcall	RXtoReset	;reset timeout
	;
	sbs	FlagVar,0
	;
	pop	R16
	out	sreg,R16
	pop	R18
	popw	R16
	reti


;************************************************************************************************************************
;	receive timeout routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	reset timeout 
;------------------------------------------------------------------------------------------------------------------------
RXtoReset:	cli
	outi	sfior,(1<<psr10)
	outiw	tcnt1,0
	outi	tifr,(1<<ocf1a)
	stsi	RXtoVar,0xFF
	reti
;------------------------------------------------------------------------------------------------------------------------	
;	timer 1 compare match interruption routine - timeout
;------------------------------------------------------------------------------------------------------------------------
RXtimeout:	push	R16
	in	R16,sreg
	push	R16
	;
	stsi	RXtoVar,0
	;
	pop	R16
	out	sreg,R16
	pop	R16
	reti
	
	
	
;************************************************************************************************************************
;	math routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------
; div24uc3 - 24/24 bit unsigned division, something between code size and speed optimized
; deleni dvou neznaminkovych 24-bit cisel, kompromis mezi delkou kodu a rychlosti
;
; dDd24r0u:dDd24r1u:dDd24r2u <- dDd24r0u:dDd24r1u:dDd24r2u div dDs24r0u:dDs24r1u:dDs24r2u
; dRe24r0u:dRe24r1u:dRe24r2u <- dDd24r0u:dDd24r1u:dDd24r2u mod dDs24r0u:dDs24r1u:dDs24r2u
;
; words:  40 + ret
; cycles: 325/342/359 + ret [min/avg/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------
	.def	dRe24r0u=R13
	.def	dRe24r1u=R14
	.def	dRe24r2u=R15
	.def	dDd24r0u=R16
	.def	dDd24r1u=R17
	.def	dDd24r2u=R18
	.def	dDs24r0u=R19
	.def	dDs24r1u=R20
	.def	dDs24r2u=R21
	.def	d24cntu=R22
	;
div24uc3:	clr	dRe24r0u
	clr	dRe24r1u
	sub	dRe24r2u,dRe24r2u
	ldi	d24cntu,13
d24u_1c3:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	dec	d24cntu
	breq	d24u_2c3
d24u_3c3:	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_4c3
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
d24u_4c3:	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	rol	dRe24r0u
	rol	dRe24r1u
	rol	dRe24r2u
	sub	dRe24r0u,dDs24r0u
	sbc	dRe24r1u,dDs24r1u
	sbc	dRe24r2u,dDs24r2u
	brcc	d24u_1c3
	add	dRe24r0u,dDs24r0u
	adc	dRe24r1u,dDs24r1u
	adc	dRe24r2u,dDs24r2u
	rol	dDd24r0u
	rol	dDd24r1u
	rol	dDd24r2u
	dec	d24cntu
	brne	d24u_3c3
d24u_2c3:	com	dDd24r0u
	com	dDd24r1u
	com	dDd24r2u
	ret

;------------------------------------------------------------------------------------------------------------------------
; bin2BCD16p5 - 16-bit binary to 5-digit packed BCD conversion (0..65535)
; prevod 16-bit binarniho cisla na peticiferne cislo v BCD
;
; BCD16res0:BCD16res1:BCD16res2 <- bin16r0:bin16r1
; BCD(0x01 0x23 0x45) <- bin(12345)
;      ^         ^
; BCD16res2  BCD16res0
;
; words:  29 + ret
; cycles: 29/180 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD16res0=R13
	.def	BCD16res1=R14
	.def	BCD16res2=R15
	.def	bin16r0=R16
	.def	bin16r1=R17
	.def	BB16_cnt1=R30
	.def	BB16_cnt2=R31
	;
bin2BCD16p5:	ser	BB16_cnt2
	mov	BCD16res2,BB16_cnt2
BB16p5_L10k:	inc	BCD16res2
	subi	bin16r0,low(10000)
	sbci	bin16r1,high(10000)
	brcc	BB16p5_L10k
	subi	bin16r0,low(-10000)
	sbci	bin16r1,high(-10000)
	ldi	BB16_cnt1,(256-16)
BB16p5_L1k:	subi	BB16_cnt1,(-16)
	subi	bin16r0,low(1000)
	sbci	bin16r1,high(1000)
	brcc	BB16p5_L1k
	subi	bin16r0,low(-1000)
	sbci	bin16r1,high(-1000)
	mov	BCD16res1,BB16_cnt2
BB16p5_L100:	inc	BCD16res1
	subi	bin16r0,low(100)
	sbci	bin16r1,high(100)
	brcc	BB16p5_L100
	subi	bin16r0,low(-100)
	or	BCD16res1,BB16_cnt1
	ldi	BB16_cnt1,(256-16)
BB16p5_L10:	subi	BB16_cnt1,(-16)
	subi	bin16r0,10
	brcc	BB16p5_L10
	subi	bin16r0,-10
	mov	BCD16res0,BB16_cnt1
	or	BCD16res0,bin16r0
	ret
		


;************************************************************************************************************************
; pseudo DMA LCD routines
; using some routines from direct HD44780 controlling routines
;  
; author: Stanislav Maslan, s.maslan@seznam.cz
;************************************************************************************************************************
;************************************************************************************************************************
;	LCD DMA refresh interruption routine (called by timer 0 compare match) 
;************************************************************************************************************************	
INTLCDDMArefresh:	pushw	R16
	pushw	R28
	in	R16,sreg
	push	R16
	;
	cbii	timsk,ocie0
	outi	tifr,ocf0
	sei
	;
	lds	R16,LCDDMAflags
	rjrc	R16,DposReset,INTLCDDMArefNNL
	;reset HD44780 cursor pos
	cbr	R16,(1<<DposReset)
	sts	LCDDMAflags,R16	;clear flag
	;reset pos
	cbi	LCDportRS,LRS	;instruction
	ldi	R16,0b10000000
	rcall	LCDbyteWriteI
	rjmp	INTLCDDMArefNL
	;no cursor reset
INTLCDDMArefNNL:	rjrc	R16,Dpos2ln,INTLCDDMArefNSL
	;break first line
	cbr	R16,(1<<Dpos2ln)
	sts	LCDDMAflags,R16	;clear flag
	;
	cbi	LCDportRS,LRS	;instruction
	ldi	R16,0b10000000+0x40
	rcall	LCDbyteWriteI
	rjmp	INTLCDDMArefNL
INTLCDDMArefNSL:	sbi	LCDportRS,LRS	;data
	ldsw	R28,LCDDMAdataBufPos
	ld	R16,Y+
	;high nibble
	rcall	LCDbyteWriteI
	;
	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen
	brne	LCDDMAnslni
	;break line
	sbs	LCDDMAflags,Dpos2ln
	rjmp	INTLCDDMArefNC
LCDDMAnslni:	cpiwh	R28,R16,LCDDMAdataBuf+LCDlineLen*LCDlinesCnt
	brlo	INTLCDDMArefNC
	;refresh completed
	stsi	LCDDMAflags,0	;clear all flags
	cbii	timsk,ocie0	;disable refresh INT
	stsw	LCDDMAdataBufPos,R28
	rjmp	INTLCDDMArefNC1	;exit
	;refresh not completed
INTLCDDMArefNC:	stsw	LCDDMAdataBufPos,R28
INTLCDDMArefNL:	;reset timer
	cli
	sbii	timsk,ocie0
INTLCDDMArefNC1:	outi	tcnt0,0
	outi	tifr,(1<<ocf0)
	outi	sfior,(1<<psr10)
	;
	pop	R16
	out	sreg,R16
	popw	R28
	popw	R16
	reti
	;write byte to LCD in 4-bit mode
LCDbyteWriteI:	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	ret
;========================================================================================================================
;	Initialize LCD DMA refresh for R16 (1 or 2) lines LCD
;	byte TX period T=50us (don't forget set XTAL freq.!!!)
;========================================================================================================================
	;some LCD parameters
	.equ	LCDlineLen=20
	.equ	LCDlinesCnt=2
	;DMA flag byte bits
	.equ	DrefreshRequest=0
	.equ	DrefreshInProgress=1
	.equ	DposReset=2
	.equ	Dpos2ln=3
	;LCD flag byte bits
	.equ	DautoRefresh=0

	
	;some variables for DMA refreshing
	.dseg
LCDDMAflags:	.byte	1	;LCD DMA refresh bit flags
LCDflags:	.byte	1	;LCD bit flags
LCDDMAdataBuf:	.byte	LCDlineLen*LCDlinesCnt	;character data buffer
LCDdataBufPtr:	.byte	2	;write buffer poitner
LCDDMAdataBufPos:	.byte	2	;buffer position
	.cseg

	;main init routine
Dinit:	rcall	LCDinit	;init LCD
	;fill entire character buffer with ' '
	rcall	DclrAll
	
	stsi	LCDflags,0	;clear all LCD bit flags
	stsi	LCDDMAflags,0	;clear refresh flags
	stsiw	LCDDMAdataBufPos,LCDDMAdataBuf	;reset buffer position
		
	outi	tccr0,(1<<cs01)|(1<<ctc0)	;timer 0 XTAL/8, CTC
	outi	ocr0,XTAL/8/20000-1	;f=20kHz (T=50us)
	ret
;========================================================================================================================
;	start LCD DMA refresh
;========================================================================================================================	
Drefresh:	in	R0,sreg
	cli
	outi	tcnt0,0
	outi	tifr,(1<<ocf0)
	outi	sfior,(1<<psr10)	;reset timer
	sbii	timsk,ocie0	;enable refresh IRQ
	stsiw	LCDDMAdataBufPos,LCDDMAdataBuf	;reset buffer position
	stsi	LCDDMAflags,(1<<DrefreshRequest)|(1<<DposReset)
	out	sreg,R0
	ret
;========================================================================================================================
;	wait for LCD DMA refresh complete 
;========================================================================================================================
DwaitRefresh:	lds	R16,LCDDMAflags
	andi	R16,(1<<DrefreshRequest)|(1<<DrefreshInProgress)
	brne	DwaitRefresh	;wait if there is request for refresh or refresh in progress
	ret
;========================================================================================================================
;	test for LCD DMA refresh complete (T=1 if in progress (or i there is request for that), T=0 if done) 
;========================================================================================================================
DtestRefresh:	set		;in progress
	lds	R16,LCDDMAflags
	andi	R16,(1<<DrefreshRequest)|(1<<DrefreshInProgress)
	brne	PC+2
	clt		;done
	ret
;========================================================================================================================
;	enable auto refresh
;========================================================================================================================
DautoRefreshSet:	sbs	LCDflags,DautoRefresh
	ret
;========================================================================================================================
;	disable auto refresh
;========================================================================================================================
DautoRefreshClr:	cbs	LCDflags,DautoRefresh
	ret
;========================================================================================================================
;	write R16 char
;========================================================================================================================
DwrChar:	ldsw	R28,LCDdataBufPtr
	st	Y+,R16
	stsw	LCDdataBufPtr,R28
	;test autorefresh
	lds	R16,LCDflags
	rjrc	R16,DautoRefresh,PC+2
	rcall	Drefresh	;start refresh if its allowed
	ret
;========================================================================================================================
;	write Z pointed FLASH string into LCD data buffer 
;========================================================================================================================
DwrStrFLASH:	ldsw	R28,LCDdataBufPtr	;load cursor pos
	rjmp	DwrStrFLASHs
DwrStrFLASHL:	st	Y+,R16	;write it to the LCD buffer
DwrStrFLASHs:	lpm	R16,Z+	;red char from FLASH
	cpi	R16,0	;break line?
	brne	DwrStrFLASHb
	ldiw	R28,LCDDMAdataBuf+LCDlineLen	;next line
	rjmp	DwrStrFLASHs
DwrStrFLASHb:	cpi	R16,0xFF	;string end?
	brne	DwrStrFLASHL
	stsw	LCDdataBufPtr,R28	;store cursor pos
	;test autorefresh
	lds	R16,LCDflags
	rjrc	R16,DautoRefresh,PC+2
	rcall	Drefresh	;start refresh if its allowed	
	ret
;========================================================================================================================
;	set cursor pos to R16
;	if your LCD size is 2x20: 
;	R16=10 - 1.line, 10.char
;	R16=20+5 - 2.line, 5.char
;========================================================================================================================	
DsetPos:	clr	R17
	subiw	R16,-(LCDDMAdataBuf)	;add buffer offset
	stsw	LCDdataBufPtr,R16
	ret
;========================================================================================================================
;	clear entire buffer, reset cursor to 0,0 (no auto refresh after this routine)
;========================================================================================================================
DclrAll:	;erase buffer (write ' ')
	ldiw	R28,LCDDMAdataBuf
	ldi	R24,LCDlineLen*LCDlinesCnt
	ldi	R16,' '
	st	Y+,R16
	dbnz	R24,PC-2
	stsiw	LCDdataBufPtr,LCDDMAdataBuf	;write cursor to 0,0
	ret
;========================================================================================================================
;	clear R16 line, reset cursor to its start
;========================================================================================================================
DclrLine:	dec	R16
	ldi	R17,20
	mul	R16,R17
	movw	R28,R0
	subiw	R28,-(LCDDMAdataBuf)	;line start address
	stsw	LCDdataBufPtr,R28	;store cursor pos
	;clear rest of this line
	ldi	R24,LCDlineLen
	ldi	R16,' '
	st	Y+,R16
	dbnz	R24,PC-2
	ret	
;========================================================================================================================
;	write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines for bin/BCD conversion
;========================================================================================================================
	;with white leading zeroes
DwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	DwrBCD16
	;with leading zeroes
DwrBCD16z:	set
	rjmp	DwrBCD16
	;skip leading zeroes
DwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
DwrBCD16:	mov	R16,R15
	rcall	DwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	DwrCharBCD
	mov	R16,R14
	rcall	DwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	DwrCharBCD
	set
	mov	R16,R13
	rcall	DwrCharBCD
	ret
	;common write BCD num.
DwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	DwrCharBCDnz	;zero?
	brts	DwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	DwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	DwrChar	;write number
DwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	DwrChar	;write number
DwrCharBCDzc:	ret
	



;************************************************************************************************************************
; HD44780 LCD routines 4-wires interface, no RW pin used - constant delay 40us
; don't forget set correct XTAL freq somewhere
; don't forget init ddrx registers for E,RS and DATA pins immediately after MCU reset!!!
;
; author: Stanislav Maslan, s.maslan@seznam.cz 
;************************************************************************************************************************
	;pin/port definitions
	.equ	LCDportE=portd	;E pin port
	.equ	LCK=5	;pin?
	.equ	LCDportRS=portd	;RS pin port
	.equ	LRS=6	;pin?
	.equ	LCDportDATA=portc	;DATA pins port
	.equ	LCDportDATAnib=0	;low=0/high=1 nibble of the port

;------------------------------------------------------------------------------------------------------------------------
;	LCD initialization, call this routine first but generate at least 20ms delay from power on!
;	R16=1 one line, R16=2 two lines
;------------------------------------------------------------------------------------------------------------------------
LCDinit:	cbi	LCDportRS,LRS	;instructions
	push	R16
	;
	in	R0,sreg
	cli		;disable IRQ
	;back to 8-bit, if it was reset (brown out detector or ISP reset)
	ldi	R16,0b00000011	;to 8-bit mode
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	ldi	R16,0b00000000	;if it was 4-bit mode (after reset) complete instruction with low nibble 0b0000
	rcall	LCDnibOut	;if it was 8-bit mode, 0b00000000 instruction will be ignored
	rcall	LCDclock	;clock
	rcall	LCDwait164	;wait 1,64ms
	;now we are surely in 8-bit mode, init 4-bit mode
	ldi	R16,0b00000010
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	;
	out	sreg,R0
	;now select mode (R16=1 one line, R16=2 two lines)
	pop	R16
	dec	R16	;N--
	swap	R16
	lsr	R16	;N<<3
	andi	R16,(1<<3)
	ori	R16,0b00100000	;add init mask
	rcall	LCDwrInstr	;full init again into 4-bit mode, selected number of lines
	ldi	R16,0b00001100
	rcall	LCDwrInstr	;turn on display
	ldi	R16,0b00000110
	rcall	LCDwrInstr	;set cursor shift mode
	rcall	LCDclrAll	;clear all
	sbi	LCDportRS,LRS	;data	
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines
;------------------------------------------------------------------------------------------------------------------------
	;with white leading zeroes
LCDwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	LCDwrBCD16
	;with leading zeroes
LCDwrBCD16z:	set
	rjmp	LCDwrBCD16
	;skip leading zeroes
LCDwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
LCDwrBCD16:	mov	R16,R15
	rcall	LCDwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	LCDwrCharBCD
	mov	R16,R14
	rcall	LCDwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	LCDwrCharBCD
	set
	mov	R16,R13
	rcall	LCDwrCharBCD
	ret
	;common write BCD num.
LCDwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	LCDwrCharBCDnz	;zero?
	brts	LCDwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	LCDwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	LCDwrChar	;write number
LCDwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	LCDwrChar	;write number
LCDwrCharBCDzc:	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Z pointed (in bytes!) 0xFF terminated string from FLASH, 0x00 line break
;	.db "text for 1st line",0x00,"text for second line",0xFF 
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrFlash:	lpm	R16,Z+	;get char
	rcall	LCDwrChar
	lpm	R16,Z
	cpi	R16,0x00
	brne	LCDwrStrFlashNBL
	adiw	R30,1
	ldi	R16,0x40
	rcall	LCDsetPos
	rjmp	LCDwrStrFlash
LCDwrStrFlashNBL:	cpi	R16,0xFF	;str end?
	brne	LCDwrStrFlash	;next char
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Y pointed RAM array, R24=length
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrRAM:	ld	R16,Y+
	rcall	LCDwrChar
	dbnz	R24,LCDwrStrRAM
	ret
;------------------------------------------------------------------------------------------------------------------------
;	set pos (R16):
;	1.line from 0x00
;	2.line from 0x40
;	3.line from 0x14
;	4.line from 0x54
;------------------------------------------------------------------------------------------------------------------------
LCDsetPos:	ori	R16,1<<7
	rjmp	LCDwrInstr
;------------------------------------------------------------------------------------------------------------------------
;	clr all
;------------------------------------------------------------------------------------------------------------------------
LCDclrAll:	ldi	R16,0b00000001
	rcall	LCDwrInstr
	rjmp	LCDwait164
;------------------------------------------------------------------------------------------------------------------------
;	clr R16 line, set cursor left on the line
;------------------------------------------------------------------------------------------------------------------------
LCDclrLine:	cpi	R16,1
	brne	LCDclrLineNo1
	ldi	R24,40	;40 chars
	ldi	R16,0x00	;1.line
	rjmp	LCDclrLineLC
LCDclrLineNo1:	cpi	R16,2
	brne	LCDclrLineNo2
	ldi	R24,40	;40 chars
	ldi	R16,0x40	;2.line
	rjmp	LCDclrLineLC
LCDclrLineNo2:	cpi	R16,3
	brne	LCDclrLineNo3
	ldi	R24,20	;20 chars
	ldi	R16,0x14	;3.line
	rjmp	LCDclrLineLC		
LCDclrLineNo3:	ldi	R24,20	;20 chars
	ldi	R16,0x54	;4.line
	rjmp	LCDclrLineLC
	;common clr routine
LCDclrLineLC:	push	R16
	rcall	LCDsetPos
LCDclrLineL:	ldi	R16,' '
	rcall	LCDwrChar
	dbnz	R24,LCDclrLineL
	pop	R16
	rcall	LCDsetPos
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 40us
;------------------------------------------------------------------------------------------------------------------------	
LCDwait40:	ldi	R16,XTAL/74000
LCDwait40_1:	dbnz	R16,LCDwait40_1
	ret
;------------------------------------------------------------------------------------------------------------------------
;	wait 1.64ms
;------------------------------------------------------------------------------------------------------------------------	
LCDwait164:	ldi	R17,41	;41*40us
LCDwait164_2:	ldi	R16,XTAL/74000
LCDwait164_1:	dbnz	R16,LCDwait164_1
	dbnz	R17,LCDwait164_2
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write R16 char, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrChar:	sbi	LCDportRS,LRS	;data
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 instr, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrInstr:	cbi	LCDportRS,LRS	;instructions
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 byte
;------------------------------------------------------------------------------------------------------------------------
LCDwrByte:	in	R0,sreg
	cli		;disable IRQ
	;high nibble
	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	out	sreg,R0
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write low nibble R16 to the LCD data port
;------------------------------------------------------------------------------------------------------------------------
LCDnibOut:	andi	R16,0b00001111	;mask low nibble
	.if	LCDportDATAnib==0	;low nibble?
	;low nibble
	in	R17,LCDportDATA
	andi	R17,0b11110000
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.elif	LCDportDATAnib==1	;high nibble?
	;high nibble
	swap	R16
	in	R17,LCDportDATA
	andi	R17,0b00001111
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.else
	.error "low=0 or high=1 nibble for LCD???"
	.endif
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD clock
;------------------------------------------------------------------------------------------------------------------------	
LCDclock:	sbi	LCDportE,LCK
	nop
	nop
	cbi	LCDportE,LCK
	ret



;************************************************************************************************************************
;	delay routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	variable delay (R16 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayVar:	pushw	R18
DelayVar_3:	ldi	R19,250
DelayVar_2:	ldi	R18,XTAL/74000
DelayVar_1:	dbnz	R18,DelayVar_1
	dbnz	R19,DelayVar_2
	dbnz	R16,DelayVar_3
	popw	R18
	ret
