;compiler: avrasm

;************************************************************************************************************************
;************************************************************************************************************************
;** rotary encoder demo, double precision - only for optical encoders
;** ukazkovy program pro praci s rotacnim enkoderem, dvojita presnost - jen pro opticke enkodery (bez aretace)
;************************************************************************************************************************
;************************************************************************************************************************ 

	.include "m32def.inc"
	.include "macros.asm"

	.equ	XTAL=16000000	;clock freq in Hz
	
	;portd pins
	.equ	RE0_ck=2	;rotary encoder 1 clock pin	
	.equ	RE0_dt=3	;rotary encoder 1 data pin

	;RAM variables
	.dseg
SomeVariable:	.byte	2	;our 16-bit var controlled by encoder (nase 16bit promenna ovlivnovana enkoderem)	
	
	;code
	.cseg
	.org	0x0000
	rjmp	Init	;prog. start
	.org	0x0002
	rjmp	INTrotEnc0	;rotary encoder 0 INT
	
;************************************************************************************************************************
;	rotary encoder INT (preruseni rotacniho enkoderu)
;************************************************************************************************************************	
INTrotEnc0:	pushw	R16
	in	R16,sreg
	push	R16
	;
	ldsw	R16,SomeVariable	;load our var (nacist nasi promennou)
	
	sbic	pind,RE0_ck	;falling or rising edge INT?
	rjmp	INTrotEnc0_1
	;falling edge	
	sbis	pind,RE0_dt	;data pin low or high? (data pin high nebo low?)
	rjmp	INTrotEnc0_inc	
	;inc var
	rcall	VarInc
	rjmp	INTrotEnc0_dec
	;dec var
INTrotEnc0_inc:	rcall	VarDec	
INTrotEnc0_dec:	rjmp	INTrotEnc0done
	
	;rising edge
INTrotEnc0_1:	sbic	pind,RE0_dt	;data pin low or high? (data pin high nebo low?)
	rjmp	INTrotEnc0_inc1	
	;inc var
	rcall	VarInc
	rjmp	INTrotEnc0_dec1
	;dec var
INTrotEnc0_inc1:	rcall	VarDec	
INTrotEnc0_dec1:	
INTrotEnc0done:	stsw	SomeVariable,R16	;store it back (a ulozit zmenenou hodnotu zpet)
	;now clear INT0 IRQ request if there were some fake requests
	;nulovat priznak preruseni pokud byly nejake vetsi zakmity
	outi	gifr,(1<<intf0)	
	;
	pop	R16
	out	sreg,R16
	popw	R16
	reti
;------------------------------------------------------------------------------------------------------------------------	
;	dec R16:R17
;------------------------------------------------------------------------------------------------------------------------
VarDec:	subiw	R16,1	;dec
	brcc	PC+3
	ldiw	R16,0	;if it's negativ (underflow), set to 0 (omezit rozsah na 0 (kvuli podteceni))
	ret	
;------------------------------------------------------------------------------------------------------------------------
;	inc R16:R17
;------------------------------------------------------------------------------------------------------------------------
VarInc:	subiw	R16,-1	;inc
	brcs	PC+3
	ldiw	R16,65536-1	;if there is overflow, set max val (omezit rozsah do 65535 (preteceni))
	ret	
;************************************************************************************************************************
;	prog. start
;************************************************************************************************************************	
Init:	outi	ddrc,0b00001111	;LCD data port
	outi	ddrd,(1<<LCK)|(1<<LRS)	;LCD ctrl pins
	
	outi	portd,(1<<RE0_ck)|(1<<RE0_dt)	;rotary encoder pull-up resistors
	
	cbi	portd,LCK	;LCD E pin = 0 (don't forget this!)
	
	ldiw	R16,ramend
	outw	sp,R16	;set stack
	
	;external IRQ for rotary encoder
	;externi preruseni pro rotacni enkoder
	outi	mcucr,(1<<isc00)	;INT0 logical change interruption (preruseni na zmenu log urovne na INT0)
	outi	gicr,(1<<int0)	;enable INT0 (povolit INT0 preruseni)
	
	
	ldiw	R16,1
	stsw	SomeVariable,R16	;clear our variable (nulovat nasi promennou)
	
	sei		;enable IRQ (povolit preruseni)
	
	ldi	R16,2
	rcall	DelayVar	;wait 20ms
	
	ldi	R16,2
	rcall	LCDinit	;init LCD as 2 lines	
	
	;main prog. loop (hlavni programova smycka)	
Loop:	ldi	R16,0
	rcall	LCDsetPos	;back to 1.line (zpet na prvni radek)
	
	ldiw	R30,Txt1T*2
	rcall	LCDwrStrFLASH	;"X="
	
	cli
	ldsw	R16,SomeVariable	;read our var (precist stav nasi promenne)
	sei
	
	rcall	bin2BCD16p5	;to BCD (prevest na BCD cislo)
	
	rcall	LCDwrBCD16wz	;write to LCD (vypsat cislo na LCD)
	
	ldi	R16,5	;wait 50ms
	rcall	DelayVar
	
	rjmp	Loop
	
	
	;text strings
Txt1T:	.db	"X=",0xFF





;************************************************************************************************************************
;	delay routines
;************************************************************************************************************************
;------------------------------------------------------------------------------------------------------------------------
;	variable long delay (R16:R17 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayLongVar:	pushw	R18
DelayLongVar_3:	ldi	R19,250
DelayLongVar_2:	ldi	R18,XTAL/74000
DelayLongVar_1:	dbnz	R18,DelayLongVar_1
	dbnz	R19,DelayLongVar_2
	dbnzw	R16,DelayLongVar_3
	popw	R18
	ret
;------------------------------------------------------------------------------------------------------------------------
;	variable delay (R16 * 10ms)
;------------------------------------------------------------------------------------------------------------------------
DelayVar:	pushw	R18
DelayVar_3:	ldi	R19,250
DelayVar_2:	ldi	R18,XTAL/74000
DelayVar_1:	dbnz	R18,DelayVar_1
	dbnz	R19,DelayVar_2
	dbnz	R16,DelayVar_3
	popw	R18
	ret
;************************************************************************************************************************
;	math routines
;************************************************************************************************************************	
;------------------------------------------------------------------------------------------------------------------------
; bin2BCD16p5 - 16-bit binary to 5-digit packed BCD conversion (0..65535)
; prevod 16-bit binarniho cisla na peticiferne cislo v BCD
;
; BCD16res0:BCD16res1:BCD16res2 <- bin16r0:bin16r1
; BCD(0x01 0x23 0x45) <- bin(12345)
;      ^         ^
; BCD16res2  BCD16res0
;
; words:  29 + ret
; cycles: 29/180 + ret [min/max]
;
; author: Stanislav Maslan, s.maslan@seznam.cz
;------------------------------------------------------------------------------------------------------------------------
	.def	BCD16res0=R13
	.def	BCD16res1=R14
	.def	BCD16res2=R15
	.def	bin16r0=R16
	.def	bin16r1=R17
	.def	BB16_cnt1=R30
	.def	BB16_cnt2=R31
	;
bin2BCD16p5:	ser	BB16_cnt2
	mov	BCD16res2,BB16_cnt2
BB16p5_L10k:	inc	BCD16res2
	subi	bin16r0,low(10000)
	sbci	bin16r1,high(10000)
	brcc	BB16p5_L10k
	subi	bin16r0,low(-10000)
	sbci	bin16r1,high(-10000)
	ldi	BB16_cnt1,(256-16)
BB16p5_L1k:	subi	BB16_cnt1,(-16)
	subi	bin16r0,low(1000)
	sbci	bin16r1,high(1000)
	brcc	BB16p5_L1k
	subi	bin16r0,low(-1000)
	sbci	bin16r1,high(-1000)
	mov	BCD16res1,BB16_cnt2
BB16p5_L100:	inc	BCD16res1
	subi	bin16r0,low(100)
	sbci	bin16r1,high(100)
	brcc	BB16p5_L100
	subi	bin16r0,low(-100)
	or	BCD16res1,BB16_cnt1
	ldi	BB16_cnt1,(256-16)
BB16p5_L10:	subi	BB16_cnt1,(-16)
	subi	bin16r0,10
	brcc	BB16p5_L10
	subi	bin16r0,-10
	mov	BCD16res0,BB16_cnt1
	or	BCD16res0,bin16r0
	ret
;************************************************************************************************************************
; HD44780 LCD routines 4-wires interface, no RW pin used - constant delay 40us
; don't forget set correct XTAL freq somewhere
; don't forget init ddrx registers for E,RS and DATA pins immediately after MCU reset!!!
;
; author: Stanislav Maslan, s.maslan@seznam.cz 
;************************************************************************************************************************
	
	;pin/port definitions
	.equ	LCDportE=portd	;E pin port
	.equ	LCK=5	;pin?
	.equ	LCDportRS=portd	;RS pin port
	.equ	LRS=6	;pin?
	.equ	LCDportDATA=portc	;DATA pins port
	.equ	LCDportDATAnib=0	;low=0/high=1 nibble of the port

;------------------------------------------------------------------------------------------------------------------------
;	LCD initialization, call this routine first but generate at least 20ms delay from power on!
;	R16=1 one line, R16=2 two lines
;------------------------------------------------------------------------------------------------------------------------
LCDinit:	cbi	LCDportRS,LRS	;instructions
	push	R16
	;
	in	R0,sreg
	cli		;disable IRQ
	;back to 8-bit, if it was reset (brown out detector or ISP reset)
	ldi	R16,0b00000011	;to 8-bit mode
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	ldi	R16,0b00000000	;if it was 4-bit mode (after reset) complete instruction with low nibble 0b0000
	rcall	LCDnibOut	;if it was 8-bit mode, 0b00000000 instruction will be ignored
	rcall	LCDclock	;clock
	rcall	LCDwait164	;wait 1,64ms
	;now we are surely in 8-bit mode, init 4-bit mode
	ldi	R16,0b00000010
	rcall	LCDnibOut
	rcall	LCDclock	;clock
	rcall	LCDwait40	;wait 40us
	;
	out	sreg,R0
	;now select mode (R16=1 one line, R16=2 two lines)
	pop	R16
	dec	R16	;N--
	swap	R16
	lsr	R16	;N<<3
	andi	R16,(1<<3)
	ori	R16,0b00100000	;add init mask
	rcall	LCDwrInstr	;full init again into 4-bit mode, selected number of lines
	ldi	R16,0b00001100
	rcall	LCDwrInstr	;turn on display
	ldi	R16,0b00000110
	rcall	LCDwrInstr	;set cursor shift mode
	rjmp	LCDclrAll	;clear all
;------------------------------------------------------------------------------------------------------------------------
;	LCD write 16-bit BCD from R13:R14:R15 (packed format), use my speed optimized routines
;------------------------------------------------------------------------------------------------------------------------
	;with white leading zeroes
LCDwrBCD16wz:	clt
	clr	R0
	com	R0
	rjmp	LCDwrBCD16
	;with leading zeroes
LCDwrBCD16z:	set
	rjmp	LCDwrBCD16
	;skip leading zeroes
LCDwrBCD16sz:	clr	R0
	clt
	;common routine, don't call this directly!
LCDwrBCD16:	mov	R16,R15
	rcall	LCDwrCharBCD
	mov	R16,R14
	swap	R16
	rcall	LCDwrCharBCD
	mov	R16,R14
	rcall	LCDwrCharBCD
	mov	R16,R13
	swap	R16
	rcall	LCDwrCharBCD
	set
	mov	R16,R13
	rcall	LCDwrCharBCD
	ret
	;common write BCD num.
LCDwrCharBCD:	andi	R16,0b00001111	;mask low nibble
	brne	LCDwrCharBCDnz	;zero?
	brts	LCDwrCharBCDnz	;leading zero?
	;leading zero
	tst	R0	;space or nothing?
	breq	LCDwrCharBCDzc	;nothing
	ldi	R16,' '
	rjmp	LCDwrChar	;write number
LCDwrCharBCDnz:	subi	R16,-'0'	;to ASCII
	set
	rjmp	LCDwrChar	;write number
LCDwrCharBCDzc:	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Z pointed (in bytes!) 0xFF terminated string from FLASH, 0x00 line break
;	.db "text for 1st line",0x00,"text for second line",0xFF 
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrFlash:	lpm	R16,Z+	;get char
	rcall	LCDwrChar
	lpm	R16,Z
	cpi	R16,0x00
	brne	LCDwrStrFlashNBL
	adiw	R30,1
	ldi	R16,0x40
	rcall	LCDsetPos
	rjmp	LCDwrStrFlash
LCDwrStrFlashNBL:	cpi	R16,0xFF	;str end?
	brne	LCDwrStrFlash	;next char
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write Y pointed RAM array, R24=length
;------------------------------------------------------------------------------------------------------------------------
LCDwrStrRAM:	ld	R16,Y+
	rcall	LCDwrChar
	dbnz	R24,LCDwrStrRAM
	ret
;------------------------------------------------------------------------------------------------------------------------
;	set pos (R16):
;	1.line from 0x00
;	2.line from 0x40
;	3.line from 0x14
;	4.line from 0x54
;------------------------------------------------------------------------------------------------------------------------
LCDsetPos:	ori	R16,1<<7
	rjmp	LCDwrInstr
;------------------------------------------------------------------------------------------------------------------------
;	clr all
;------------------------------------------------------------------------------------------------------------------------
LCDclrAll:	ldi	R16,0b00000001
	rcall	LCDwrInstr
	rjmp	LCDwait164
;------------------------------------------------------------------------------------------------------------------------
;	clr R16 line, set cursor left on the line
;------------------------------------------------------------------------------------------------------------------------
LCDclrLine:	cpi	R16,1
	brne	LCDclrLineNo1
	ldi	R24,40	;40 chars
	ldi	R16,0x00	;1.line
	rjmp	LCDclrLineLC
LCDclrLineNo1:	cpi	R16,2
	brne	LCDclrLineNo2
	ldi	R24,40	;40 chars
	ldi	R16,0x40	;2.line
	rjmp	LCDclrLineLC
LCDclrLineNo2:	cpi	R16,3
	brne	LCDclrLineNo3
	ldi	R24,20	;20 chars
	ldi	R16,0x14	;3.line
	rjmp	LCDclrLineLC		
LCDclrLineNo3:	ldi	R24,20	;20 chars
	ldi	R16,0x54	;4.line
	rjmp	LCDclrLineLC
	;common clr routine
LCDclrLineLC:	push	R16
	rcall	LCDsetPos
LCDclrLineL:	ldi	R16,' '
	rcall	LCDwrChar
	dbnz	R24,LCDclrLineL
	pop	R16
	rcall	LCDsetPos
	ret

;------------------------------------------------------------------------------------------------------------------------
;	wait 40us
;------------------------------------------------------------------------------------------------------------------------	
LCDwait40:	ldi	R16,XTAL/74000
LCDwait40_1:	dbnz	R16,LCDwait40_1
	ret
;------------------------------------------------------------------------------------------------------------------------
;	wait 1.64ms
;------------------------------------------------------------------------------------------------------------------------	
LCDwait164:	ldi	R17,41	;41*40us
LCDwait164_2:	ldi	R16,XTAL/74000
LCDwait164_1:	dbnz	R16,LCDwait164_1
	dbnz	R17,LCDwait164_2
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write R16 char, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrChar:	sbi	LCDportRS,LRS	;data
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 instr, wait 40us
;------------------------------------------------------------------------------------------------------------------------
LCDwrInstr:	cbi	LCDportRS,LRS	;instructions
	rcall	LCDwrByte
	rjmp	LCDwait40
;------------------------------------------------------------------------------------------------------------------------
;	write R16 byte
;------------------------------------------------------------------------------------------------------------------------
LCDwrByte:	in	R0,sreg
	cli		;disable IRQ
	;high nibble
	push	R16
	swap	R16
	rcall	LCDnibOut
	rcall	LCDclock
	;low nibble
	pop	R16
	rcall	LCDnibOut
	rcall	LCDclock
	out	sreg,R0
	ret
;------------------------------------------------------------------------------------------------------------------------
;	write low nibble R16 to the LCD data port
;------------------------------------------------------------------------------------------------------------------------
LCDnibOut:	andi	R16,0b00001111	;mask low nibble
	.if	LCDportDATAnib==0	;low nibble?
	;low nibble
	in	R17,LCDportDATA
	andi	R17,0b11110000
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.elif	LCDportDATAnib==1	;high nibble?
	;high nibble
	swap	R16
	in	R17,LCDportDATA
	andi	R17,0b00001111
	or	R17,R16	;to the port
	out	LCDportDATA,R17
	.else
	.error "low=0 or high=1 nibble for LCD???"
	.endif
	ret
;------------------------------------------------------------------------------------------------------------------------
;	LCD clock
;------------------------------------------------------------------------------------------------------------------------	
LCDclock:	sbi	LCDportE,LCK
	nop
	nop
	cbi	LCDportE,LCK
	ret